(*
  Author: Prodromos Gerakios 
  All rights reserved.
 *)

open Gc_stats
open Callg
open Readcalls
open Scc
open Fstructs
open Cilinfos
open Stdutil
open Lvals
open Scope
open Fstructs

module D = Cildump
module PTA = Myptranal
module A = Alias
module SPTR = Symstate2
module SS = Symsummary
module FC = Filecache
(*module L = Logging*)
(*module RS = Racesummary*)
module BS = Backed_summary

external inc_hash :  int -> int -> int  = "inc_hash"
(******************************************************)
module E = Errormsg
open Cil
open Printf
open Pretty
(*************************************************************************** 
 *                   Definitions
 *
 *
 *
 *
 ***************************************************************************)
exception Not_implemented of string 
let rs_not_implemented format =
 Printf.ksprintf (fun x -> raise (Not_implemented x)) format
exception Impossible of string
let rs_impossible format =
 Printf.ksprintf (fun x -> raise (Impossible x)) format


 (* A context (ctx) is a list of 
  * function descriptors and a 
  * location from which the 
  * function has been called. 
  *
  * Example: 
  * (main,locUnknown)->(foo,loc_main_A)-> ....
  *
  * In the above example foo has been called in
  * main at location "loc_main_A"
  * *)
 type ctx  = (Cil.fundec * Cil.location) list

 (* a variable is uniquely identified by 
  * varinfo and the contextual information 
  * *)
 type uvid = ctx * Cil.varinfo


type sum_typ =  (Lvals.aLval * Scope.scope) list
type vinfo = Cil.fundec * Cil.varinfo 

type sexp = (* simplified abstract expression*)
 | SBot
 | SAddrOf of sexp 
 | SVar of (vinfo * Cil.offset)
 | SDeref of (sexp * Cil.offset)
 | SAbsHost of (AT.ptaNode * Cil.offset)
 (*instantiation of a variable with a list of
  * possible expressions 
  * at a specific location
  * *)
 | SInst of (vinfo * Cil.offset * sexp list * Cil.location ) 


(* location of call 
 * function expr 
 * arg expr (single arg)
 * *)
type lockopinfo = Cil.location * Cil.exp * Cil.exp

(* location of call 
 * location of function being called *)
type callinfo = Cil.location * Cil.location * Cil.varinfo

type resolved = 
 | Resolved of effect
 | Unresolved of int (* function vid. For mutually rec funs*)
and 
 atomic_effect = 
   LockOp of  
    (* location of call, location of statement, 
     * function that is called
     * argument to lock op 
     * bool flag acquire/release 
     * (aExp = aliased formal param or allocation point or global)
     * *)
   (lockopinfo * bool * sexp)

| Join  of effect list 

(*|  Branch of (effect * effect)*)
 (* call info and a reference to a resolved type
  *  - the reference allows us to perform in-place
  *    updates when resolving an effect
  * *)
| NCall of (callinfo * (resolved ref))
 (* recursive call
  * similarly to NCall, the reference is None
  * when the recursive effect has not been 
  * computed, otherwise it is Some (summarized effect)
  * *)
| RCall of (callinfo * (effect option ref))
and effect = atomic_effect list 
(************************************************************************)

(*************************************************************************** 
 *                      Utilities 
 *
 *
 *
 *
 ***************************************************************************)
let rec strstr haystack needle offset =
      if ((String.length needle) + offset) > 
         (String.length haystack) then  -1
        else
      if (String.sub haystack offset 
         (String.length needle)) = needle then  offset
      else if (String.contains_from haystack (offset+1) 
             (String.get needle 0))  then
       strstr haystack needle
            (String.index_from haystack (offset+1)
            (String.get needle 0))
      else -1

(************************************************************************)
(* convert exception to option*)
let absFind find_fn data_struct what =
    try Some (find_fn data_struct what)                  
    with Not_found -> None

let absHTprint key2str val2str ht =
  Hashtbl.iter 
  (fun key val_ -> 
     Printf.printf "%s\n%s\n" 
     (key2str key)
     (val2str val_)
  )  ht
(* far from perfect but does the thing*)
let abs_list2string sep val2str ls =
 match ls with
  | [] -> ""
  | hd::[] -> val2str hd
  | hd::tl ->
    (val2str hd) ^ 
    List.fold_left 
    (fun a b -> a ^ sep ^ (val2str b )) ""  tl

(* let s = List.fold_left 
  (fun a b -> a ^ sep ^ (val2str b )) ""  ls 
 in if s <> "" then  String.sub s 1 (String.length s -1)  else ""*)

let addList l_ref elt  = l_ref := (!l_ref)@[elt]
(************************************************************************)
exception Opt2some
(************************************************************************)
let any2str f e  = (sprint 80 (dprintf "%a" f e))
let exp2str     = any2str d_exp 
let stmt2str    = any2str d_stmt 
let typ2str     = any2str d_type
let loc2str loc =  any2str Cil.d_loc loc
let opt2bool = function Some _ -> true | _ -> false 
let opt2some = function Some x -> x | _ -> raise Opt2some
(************************************************************************)
let str2varinfo s = opt2some (varinfo_of_name s)
let fundec2loc f = f.svar.vdecl
let varinfo2loc vi = vi.vdecl
let varinfo2str vi = vi.vname 
let fundec2str f = f.svar.vname
let aexp2str x = any2str Lvals.d_aexp x
let alval2str x = any2str Lvals.d_alval x

let vid2str vid = (varinfo2str (Cilinfos.getVarinfo vid))

(*************************************************************************** 
 *                      Cil file stuff
 *
 *
 *
 *
 ***************************************************************************)

 (* A set of functions for retrieving the current 
  * Cil.file
  * *)
 let cilfiles = ref ([] : Cil.file list)
 let (cilfile: Cil.file option ref) = ref  (None : Cil.file option)

 let setCilFile (file:Cil.file) : unit = 
   if List.for_all (* add once*)
       (fun f-> f.fileName <> file.fileName) !cilfiles then
    cilfiles := file :: (!cilfiles); 

    cilfile := Some file

 let vid2fundec (id:int) : Cil.fundec = 
    match !cilfile with 
     | Some file -> 
        begin match Cilinfos.getFundec id file with
         | Some fd -> fd 
         | None -> 
           raise (Impossible "vid2fundec/1")
        end 
     | None -> 
           raise (Impossible "vid2fundec/2")

let vid2fundec_opt (id:int) : Cil.fundec option = 
    match !cilfile with 
     | Some file -> 
         Cilinfos.getFundec id file 
     | None -> 
         None

(*******************************************************
 *   function pointer alias analysis API               *
 *                                                     *
 *                                                     *
 *******************************************************)
type alias_map = (Cil.location,string list) Hashtbl.t
type malloc_map = (Cil.location,string) Hashtbl.t
type malloc_map_inv = (string,Cil.location) Hashtbl.t
type alias_bin_type = (malloc_map * malloc_map_inv * alias_map)

let alias_map0=ref (None:alias_map option)
let malloc_map0=ref (None:malloc_map option)
let malloc_map1=ref (None:malloc_map_inv option)

let loadFunAlias ()  = (*should be invoked at init*)
 let fname = Filename.concat "ciltrees" "alias.bin" in
 let ic = open_in fname in
 let (m,m',f) = (Marshal.from_channel ic:alias_bin_type) in
 malloc_map0 := Some m;
 malloc_map1 := Some m';
 alias_map0 := Some f;
 close_in ic

let alias_map loc = 
 begin try Hashtbl.find (opt2some !alias_map0) loc
 with Not_found -> [] end

let malloc_map loc = 
 begin try Hashtbl.find (opt2some !malloc_map0) loc
 with Not_found -> "" end

let is_malloc_global name = 
 Hashtbl.mem (opt2some !malloc_map1) name

(************************************************************************)
(*************************************************************************** 
 *                     Print functions 
 *
 *
 *
 *
 ***************************************************************************)
(* printf or logging wrapper *)
let info format =  Printf.ksprintf (fun x ->Printf.printf "Info: %s" x) format

let enable_log = ref false
let log format =  Printf.ksprintf 
                  (fun x ->
                   if !enable_log then 
                    begin
                       Printf.printf "%s" x;
                       flush stdout
                    end) format
let found_errors = ref false 
let hasErrors () = !found_errors
let err format = found_errors:=true;
Printf.ksprintf(fun x ->Printf.fprintf stderr "Error: %s\n" x;flush stderr) format
let dbg format = Printf.ksprintf (fun x ->Printf.printf "Debug: %s" x; flush
stdout) format

(*************************************************************************** 
 *                      sexp-related functions
 *
 *
 *
 *
 ***************************************************************************)

(***********************************************************************
 *
 *          Simplified Expression Functions
 * *********************************************************************)
(*let rec containsAbsHost (e:sexp) : bool = 
match e with   
 | SBot -> false 
 | SAddrOf e' -> containsAbsHost e'
 | SVar _ -> false
 | SDeref (e',_) -> containsAbsHost e' 
 | SAbsHost _ -> true 
 | SInst (_,_,e',_) ->  List.exists containsAbsHost e'*)

let rec detail_sexp (e :sexp) : string = 
let po (o:Cil.offset) = any2str (Cil.d_offset (text "")) o in
begin match e with 
 | SBot -> "SBot" 
 | SAddrOf e' -> "SAddrOf(" ^ (detail_sexp e') ^ ")"
 | SVar ((fd,vi),o) -> 
     "SVar(" ^ (fundec2str fd) ^ ","   ^ 
     (varinfo2str vi) ^","^(po o) ^")"
 | SDeref (e',o) -> 
     "SDeref(" ^ (detail_sexp e') ^ "," ^ 
     (po o) ^   ")"
 | SAbsHost (n,o) -> "SAbsHost(" ^
    (A.Abs.string_of n) ^ "," ^ (po o) ^ ")" 
 (*instantiation of a variable with a list of
  * possible expressions *)
 | SInst ((fd,vi),o,el,loc) ->
     "SInst(" ^ (fundec2str fd) ^ ","   ^ 
     (varinfo2str vi) ^ "," ^  (po o) ^
     "@line:" ^ (string_of_int loc.line)
      ^" ,{"^
     (abs_list2string "," detail_sexp el) ^ "})"
end 

let rec sexp2str (e :sexp) : string = 
let po (o:Cil.offset) = any2str (Cil.d_offset (text "")) o in
begin match e with 
 | SBot -> "bot" 
 | SAddrOf e' -> "&" ^ (sexp2str e')  
 | SVar ((fd,vi),o) -> 
       (fundec2str fd) ^"::" ^(loc2str vi.vdecl)^ "::" ^ (varinfo2str vi)  ^(po o) 
 | SDeref (e',o) -> 
     "*(" ^ (sexp2str e') ^  (po o) ^   ")"
 | SAbsHost (n,o) -> "#(" ^
    (A.Abs.string_of n) 
    ^ "," ^ (po o) ^ ")" 
 (*instantiation of a variable with a list of
  * possible expressions *)
 | SInst ((fd,vi),o,el,loc) ->
     "$(" ^ (fundec2str fd) ^ "::"   ^ 
     (varinfo2str vi)  ^(po o)^ "@line:" ^ 
     (string_of_int loc.line)   ^ ",{" ^
     (abs_list2string "," sexp2str el) ^ "})"
end 

let rec simplify_aExp 
(f: Cil.varinfo -> Cil.fundec) (e: Lvals.aExp) : sexp = 
let do_e = simplify_aExp  f in
let do_l ((h,o): Lvals.aLval) : sexp = 
begin match h with 
 | CVar vi -> SVar ((f vi,vi),o)
 | CMem e -> SDeref (do_e e,o)
 | AbsHost a -> SAbsHost (a,o)
end in
begin match e with 
 | CConst  _  
 | CAlignOf  _  
 | CSizeOfStr _
 | CSizeOfE _
 | CAlignOfE _
 | CUnOp _  
 | CBinOp _ 
 | CSizeOf   _ -> SBot
 | CCastE (_,e') -> do_e e'
 | CStartOf l -> do_l l 
 | CAddrOf l ->  SAddrOf (do_l l)
 | CLval l -> do_l l 
end


type subst_typ0 = (Cil.varinfo * sexp list) list 
type subst_retyp0 = sexp 

let substVarinfo_sexp 
(call_loc:Cil.location)
(s:subst_typ0) (e : sexp):subst_retyp0=
let rec do_e (e : sexp):subst_retyp0 =
begin match e with    
 | SBot -> SBot 
 | SAddrOf e' -> SAddrOf (do_e e')
 | SVar((fd,vi),o) ->
   let fnd (vi',_) = vi'.vid = vi.vid in
   begin match absFind List.find fnd s with
    | Some (_,e') -> SInst((fd,vi),o,e',call_loc) (* instantiated variable*)
    | None -> e    
   end
 | SDeref(e',o) -> SDeref (do_e e',o)
 | SAbsHost(p,o) -> e
 | SInst(x,o,e',loc) -> SInst(x,o,List.map do_e e',loc)
end 
in do_e e 

let rec singleAlias (e : sexp ) : bool = 
begin match e with    
 | SBot -> true
 | SAddrOf e' -> singleAlias e' 
 | SVar _ -> true
 | SDeref(e',_) -> singleAlias e' 
 | SAbsHost _  -> true 
 | SInst(_,_,e',_) -> (List.length e') < 2 
end 

let rec vi_of_sexp (e:sexp): varinfo = 
let rs s = raise (Impossible ("vi_of_sexp/" ^ s)) in
begin match e with    
 | SBot ->  rs "SBot"
 | SAbsHost (pta,o)  -> 
       rs_impossible "vi_of_sexp/AbsHost: %s"
       (abs_list2string ","
         (fun x -> exp2str (Lval  (addOffsetLval o (x))))
         (A.Abs.represents pta))
       (*rs "SAbsHost" *)
 | SAddrOf e' -> vi_of_sexp e' 
 | SVar ((fd,vi),_) -> vi
 | SDeref(e',_) -> vi_of_sexp e' 
 | SInst(_,_,e'::[],_) -> vi_of_sexp e'
 | _ -> rs "vi_of_sexp/SInst"  
end 

let vid_of_sexp (e:sexp): int = 
 begin  try
     (vi_of_sexp e).vid
  with 
   | e0 ->
      log "Exn from vid_of_sexp : %s\n" (sexp2str e);
     raise e0
 end

let sexp2exp (e:sexp)  : exp = 
let rec sexp2exp0 (e:sexp) (oo :offset) : exp = 
let rs s = 
  raise (Impossible ("sexp2exp0/" ^ s))  in
begin match e with    
 | SBot ->  
   rs "1"
 | SAbsHost _  ->(*this (ptanode,offset)*) 
   rs "2" 
 | SAddrOf e' ->
   begin  match sexp2exp0 e' oo with
    | Lval l ->  mkAddrOf l
    | _ -> rs "3" 
   end 
 | SVar ((fd,vi),o) -> 
   Lval (Var vi,addOffset o oo) 
 | SDeref(e',o) ->
   Lval (mkMem (sexp2exp0 e' oo) o)
 | SInst(_,o,e'::[],_) -> 
       sexp2exp0 e' (addOffset o oo) 
(*   begin  match sexp2exp0 e' with
    | Lval lv ->   Lval (addOffsetLval o lv)
    | Deref (Lval lv 
    | se ->
         rs_impossible "sexp2exp0/4 : %s\n" (exp2str se)
   end *)
 | _ -> 
     raise (Impossible "sexp2exp0/3") 
end
in sexp2exp0 e NoOffset


(* returns vid + (as the previous function)
 * AND the instantiation path (context-sensitive)
 * *)
let rec uvid_of_sexp (e:sexp): uvid = 
let rs s = raise (Impossible ("uvid_of_sexp/" ^ s)) in
begin match e with    
 | SBot ->  rs "SBot"
 | SAbsHost _  -> rs "SAbsHost" 
 | SAddrOf e' -> uvid_of_sexp e' 
 | SVar ((fd,vi),_) -> ((fd,vi.vdecl)::[],vi)
 | SDeref(e',_) -> uvid_of_sexp e' 
 | SInst((fd,_),_,e'::[],loc) -> 
   let (a,b) = uvid_of_sexp e' in
   (a@[(fd,loc)],b) 
 | _ -> rs "uvid_of_sexp/SInst"  
end 

(*prefix *)
let rec ctx_leq (l:ctx) (l':ctx) : bool =
 match (l,l') with
 | ([],_) -> 
       true 
 | ((fd1,loc1)::tl,(fd2,loc2)::tl') ->
       if fd1.svar.vid <> fd2.svar.vid ||
          loc1 <> loc2 then false
       else ctx_leq tl tl' 
 | _ -> 
       false

let rec ctx_eq (l:ctx) (l':ctx) : bool =
 match (l,l') with
 | ([],[]) -> 
       true 
 | ((fd1,loc1)::tl,(fd2,loc2)::tl') ->
       if fd1.svar.vid <> fd2.svar.vid ||
          loc1 <> loc2 then false
       else ctx_leq tl tl' 
 | _ -> 
       false


let rec sexp_leq (e0:sexp) (e1:sexp): bool =
 let ((ctx1,vi1),(ctx2,vi2))=
               (uvid_of_sexp e0,uvid_of_sexp e1) in
 vi1.vid = vi2.vid && 
 (vi1.vglob || (ctx_leq ctx1 ctx2)) 

(*
*)
  
 
let rec sexp_eq (e0:sexp) (e1:sexp): bool =
let rs s = raise (Impossible ("sexp_eq/" ^ s)) in
begin match (e0,e1) with    
 | (SBot,SBot) ->  
       true
 | (SAbsHost _,_)  -> 
       rs "SAbsHost" 
 | (_,SAbsHost _)  -> 
       rs "SAbsHost" 
 | (SAddrOf e0',SAddrOf e1')  -> 
       sexp_eq e0' e1'
 | (SVar ((fd0,vi0),_),SVar ((fd1,vi1),_)) ->
       fd0.svar.vid = fd1.svar.vid &&
       vi0.vid = vi1.vid 
 | (SDeref(e0',_),SDeref(e1',_)) ->
       sexp_eq e0' e1' 
 | (SInst((fd0,vi0),o,e0'::[],loc1),
    SInst((fd1,vi1),o',e1'::[],loc2)) -> 
    fd0.svar.vid = fd1.svar.vid  &&
    vi0.vid = vi1.vid &&
    loc1 = loc2   &&
    (sexp_eq e0' e1') &&
    (Util.equals o o')
 | _ -> 
    rs (Printf.sprintf "SInst s0=%s s1=%s"  
        (sexp2str e0) (sexp2str e1))
end 


(*************************************************************************** 
 *                      aExp/aLval -related functions 
 *
 *
 *
 *
 ***************************************************************************)
let rec print_alval (l:Lvals.aLval) : string = 
 begin match l with 
   (CVar vi , _ ) ->  "CVar(" ^ (varinfo2str vi)  ^ ")"
  |(AbsHost _ , _ ) ->  "AbsHost" 
  |(CMem e , _ ) ->  "CMem(" ^ (print_aexp e) ^")"
 end
and print_aexp (aexp:Lvals.aExp) :string = 
  let lv (l:Lvals.aLval) = "(" ^ (print_alval l) ^ ")" in
  let ae e = "(" ^ (print_aexp e) ^ ")" in
  let aee e1 e2 = "(" ^ (print_aexp e1) ^ "," ^
                   (print_aexp e2) ^ ")" in
begin match aexp with
    CConst _ -> "CConst"
  | CSizeOf  _ -> "CSizeOf"
  | CSizeOfE e  -> "CSizeOfE" ^ (ae e)
  | CSizeOfStr _ -> "CSizeOfStr" 
  | CAlignOf  _ -> "CAlignOf" 
  | CAlignOfE  e -> "CAlignOfE" ^ (ae e)
  | CUnOp (_,e,_) -> "CUnOp" ^ (ae e)
  | CBinOp (_,e1,e2,_) -> "CBinOp" ^ (aee e1 e2)
  | CCastE  _ -> "CCastE" 
  | CLval  l -> "CLval"  ^ (lv l)
  | CAddrOf l -> "CAddrOf" ^ (lv l)
  | CStartOf l -> "CStartOf"  ^ (lv l)
end  


(*************************************************************************** 
 *                     Time functions 
 *
 *
 *
 *
 ***************************************************************************)
let t_start = ref (Sys.time())
let t_sum = ref !t_start

let string_of_time (r: float) : string =
  let t = int_of_float (r *. 100.0) in
  let point = t mod 100 in
  let t = t / 100 in
  let s = t mod 60 in
  let m = (t mod 3600) / 60 in
  let h = t / 3600 in
  Printf.sprintf "%02d:%02d:%02d.%02d" h m s point
 
let start_time s = 
   log "Start time : %s\n" s; 
   t_start := Sys.time () 


let end_time s = 
  let t = (Sys.time()) -. !t_start in
  log "End time : %s  time = %s\n" s (string_of_time t);
  t_sum := !t_sum +. t

let sum_time () = 
  log "Total time: %s\n" (string_of_time !t_sum);

type time_t = {  mutable t_start : float ;
                 mutable t_sum   : float ;
              }

let rstart_new s = 
   log "Start time : %s\n" s; 
  {t_start = Sys.time () ; t_sum = 0.0 }

let rend_time s r =
 let t' = Sys.time () -. r.t_start in
 let t'' = r.t_sum +. t' in
 log "End time : %s  time=%s total=%s\n" s
 (string_of_time t') (string_of_time t'');
 r.t_start <-  Sys.time ();
 r.t_sum <-  t''

let print_mb s =
 begin
 let word_size = (Cil.bitsSizeOf (TPtr(TVoid [], [])))/8 in
 log "HEAP (MB): %d @@ %s\n"  
 (((Gc.stat ()).Gc.live_words * word_size) / 1048576 ) s
 end


(*************************************************************************** 
 *                      EFFECT SECTION 
 *
 *
 *
 *
 ***************************************************************************)
type fun_effects =  (Cil.location,
                     (Cil.fundec * effect ref * Cil.varinfo list)) Hashtbl.t

(************************************************************************)
let must_resolved (msg:string) (f: resolved ref) : effect=
 begin match !f with 
  | Resolved f' -> f'
  | Unresolved vid -> 
    raise (Impossible (msg  ^ " : Function " ^
           (varinfo2str (Cilinfos.getVarinfo vid)) ^
           " has not been resolved"))
 end 
let mkresolved (f:effect) = ref (Resolved f) 
(************************************************************************
 * Printing functions for atomic effects
 ************************************************************************)
let print_ncall_detail = ref false
let print_rcall_detail = ref false

let rec atomic_effect2str a = 
match a with
| LockOp(_,op,e)->(sexp2str  e)^(if op then " +" else " -")
| Join el -> "{" ^ 
  (abs_list2string " ? " 
  (fun z -> "(" ^ (effect2str  z) ^ ")") el) ^ "}"

(*| Branch(f1,f2) -> 
      "{("^(effect2str f1) ^ ") ? ("^(effect2str f2)^")}"*)
| NCall((_,fun_loc,_),f) -> 
   "call(" ^
   (begin match !f with
     | Resolved f' ->
       if !print_ncall_detail then (effect2str f')
       else (loc2str fun_loc)
     | Unresolved vid ->
       varinfo2str (Cilinfos.getVarinfo vid)
    end
   ) 
   ^ ")"
| RCall(_,f_opt) ->
   "rcall(" ^
   (if !print_rcall_detail then 
    begin match !f_opt with
     | Some f' -> (effect2str f')
     | None -> "unresolved" 
    end
    else "*"  
   ) ^ ")"

(*| Call_unresolved(_,_,floc)->"call(" ^ (loc2str floc) ^")"
| RecursiveCall_unresolved(loc) -> "(recursive)"
| Call_resolved (_,_,e)
| RecursiveCall_resolved e  -> effect2str e *)
(*| Iteration_resolved e -> effect2str fnd e *)
(*| Iteration_unresolved(f) -> "iteration(" ^ (effect2str fnd f) ^ ")"*)
(*| Wildcard -> "(wildcard)"*)

and
 effect2str f = 
   match f with 
    | [] -> "()" 
    | _ ->  abs_list2string "," (atomic_effect2str) f

(* control level of details when printing effects *)
let effect2str_detail (b_ncall) (b_rcall) (f:effect) = 
 let (tmp0,tmp1) = (!print_ncall_detail,
                    !print_rcall_detail)  in
 print_ncall_detail := b_ncall;
 print_rcall_detail := b_rcall;
 let str = effect2str f in
 print_ncall_detail := tmp0;
 print_rcall_detail := tmp1;
 str

let pretty_effect f = 
let rec ppeffect2str f =
 List.fold_left (fun str elt -> str ++ match elt with 
   | LockOp(_,op,e)->
     let e' = sexp2exp e in
     let s  = if op then "+" else "-" in
     text  ((exp2str e') ^ s)
   | Join el ->
     begin match el with 
     | [] -> text "{}"
     | hd::tl -> 
       (text "{") ++ (ppeffect2str hd) ++ 
        (List.fold_left (fun fst m ->
         fst ++ (text "\t?\t") ++ (ppeffect2str m) ) 
         (text "") el) 
        ++ (text "}")
     end 
   | NCall((_,fun_loc,_),f) ->
      let f' = must_resolved "ppeffect2str" f in
      (text "[ ") ++ 
      (ppeffect2str f')  ++
      (text " ]" )
   | RCall(_,_) ->
      rs_impossible "ppeffect2str/RCall"
 ) (text "") f
in ppeffect2str f

let rec printStmtGamma (fn:effect->string) (f:effect) : string = 
match f with 
  LockOp((loc,e0,e1),_,_)::tl -> (* lock operation *)
   let (s1,s2) =  ( (fn f), (printStmtGamma fn tl))
    in if s1 = "" then ""
       else  (exp2str e0) ^"(" ^ (exp2str e1)
       ^")" ^ " : "  ^  s1 ^ "\n" ^ s2 
| _::tl  ->  (printStmtGamma fn tl)
| []     -> ""

let printFunGammas (ht:fun_effects)  =  
  log "Printing statement annotations";
   absHTprint loc2str
   (fun (fdec,f,_) ->
     let tmpStr = (effect2str ) (!f)
     in if tmpStr <> "" then begin
         "========================\n" ^
        fdec.svar.vname  ^ ": " ^ tmpStr ^ "\nDetails:\n" ^
        (printStmtGamma (effect2str ) (!f)) ^ 
        "\n--------------------\n========================\n"
     end else ""
   ) ht


(*let is_formal (e:Cil.exp) = true 
let subst_args (el1 : Cil.exp list) (el2: Cil.exp list) =*)

(*a map from formals to a set of possible values *)
let rec substAtomicEffect
  (call_loc:Cil.location)
  (subst:subst_typ0) (f:atomic_effect) : atomic_effect  = 
let do_f  = substEffect call_loc subst  in
begin match f with 
 |  LockOp(calli,is_acquire,e) ->
    let e' = substVarinfo_sexp call_loc subst e in
   (*
    let e0 = (sexp2exp e') in
    let e'' = (aliases e0) in
    let e''' = List.map (simplify_aExp (fun _ -> dummyFunDec)) e'' in
    let hack = SInst((dummyFunDec,dummyFunDec.svar),NoOffset,e''',
                      locUnknown)
    in
    log "SUBST e=%s for e'%s\n" (sexp2str e) (sexp2str e');
    List.iter (fun x -> log "ALIAS::: %s  e0=%s\n HACK=%s\n\n"
                         (aexp2str x) (exp2str e0) (sexp2str hack)
              ) e'';

    begin match e''' with
    | _::[] ->  
        LockOp(calli,is_acquire,hack)
    | [] -> 
        LockOp(calli,is_acquire,e')
    | _ ->
       err "Too many aliases for LockOp var %s aliases: %s"
            (sexp2str e) (abs_list2string "," sexp2str e''')
        ; f 
    end*)
              
    if singleAlias e' then 
        LockOp(calli,is_acquire,e')
    else begin 
     err "Too many aliases for LockOp"; f 
    end

 | Join el -> 
   Join (List.map do_f el)

(* | Branch(f1,f2) -> 
   Branch (do_f f1,do_f f2)*)

 | NCall(a,f_res) -> 
   let f' = must_resolved "substAtomicEffect" f_res in
   NCall(a,mkresolved (do_f f'))

 | RCall(a,f_opt_ref) ->
   begin match !f_opt_ref with
    | Some f' -> 
          RCall(a,ref (Some (do_f f')))
    | None -> 
          raise (Impossible "substAtomicEffect/RCall")
   end
(*
 | Call_unresolved(call_loc,args0,fun_loc) -> 
   let sub = substVarinfo_sexp subst in
   let args0' = List.map (List.map sub) args0 in
   (* (fun arg->List.map sub arg)List.flatten (List.map sub arg)) args0 in*)
   Call_unresolved(call_loc,args0',fun_loc)

 | Call_resolved (l1,l2,f') -> Call_resolved(l1,l2,do_f f')
 | RecursiveCall_resolved f' -> RecursiveCall_resolved (do_f f')
 | RecursiveCall_unresolved _ -> f *)
(* | Wildcard -> f*)
(* | Iteration_unresolved f' -> Iteration_unresolved(do_f f')*)
(* Wildcard*)
(* | Iteration_resolved f' ->Iteration_resolved (do_f f')*)
end
and substEffect (call_loc:Cil.location)
(subst:subst_typ0) (f:effect):effect= 
 List.map (substAtomicEffect call_loc subst) f

let rec isFuture_sexp (e:sexp) : bool = 
begin match e with 
 | SBot -> false 
 | SAddrOf e' -> isFuture_sexp e' 
 | SVar ((fd,vi),o) -> not vi.vglob
 | SDeref (e',o) ->  isFuture_sexp e'
 | SAbsHost (n,o) -> false 
 | SInst ((fd,vi),_,el,_) -> false
   (*List.exists isFuture_sexp el*)
end 

let rec filterEffect call_loc (f: effect) : effect =
List.filter 
 (function 
  |  LockOp(calli,is_acquire,e) ->
    let b = not (isFuture_sexp e)  in
    if not b then 
       log "FILTERING %s\n" (sexp2str e);
    b
  |  _ -> true
 ) f

let rec morePreciseEffect aliases (f:effect) : effect = 
 List.fold_left
 (fun f' a -> f' @
  match a with
  | LockOp(calli,is_acquire,e) ->
    let verifyLv e = 
     let vi = vi_of_sexp e in
     let (call_loc,_,_) = calli in
     if vi.vglob = false then
       err "[%s] Lockop expression %s is not global"
          (loc2str call_loc) (exp2str (sexp2exp e))
     else  if  is_malloc_global vi.vname then
      err "[%s] Lockop expression %s is ``malloc'ed``. This feature is unsupported."
          (loc2str call_loc) (exp2str (sexp2exp e))
    in
    let e'   = sexp2exp e in
    let e''  = aliases e' in
    begin match e'' with
    | hd::[] ->  
       let e''' = simplify_aExp 
                        (fun _ -> dummyFunDec) hd in
       (*log "ALIAS::: OF %s IS %sn\n"
       (sexp2str e) (sexp2str e''');*)
       verifyLv e''';
       [LockOp(calli,is_acquire,e''')]
    | [] -> 
         (*log "SAME:::%s\n"  (exp2str e');*)
         verifyLv e;
         [a]
    | _ ->
       err "Too many aliases for LockOp var %s aliases: %s"
            (sexp2str e) 
            (abs_list2string "," aexp2str e'')
            ;[a]
    end
  | Join el -> 
    let el' = List.map 
              (fun x->morePreciseEffect aliases x) el in
    [Join el']
  | NCall(i,res) ->
    let f' = must_resolved "morePreciseEffect" res in
    let f'' = morePreciseEffect aliases f' in
    [NCall(i,mkresolved f'')]
  | RCall _ ->
    rs_impossible "morePreciseEffect\n" 
 ) [] f

(**********************************************************
 *
 *        Effect Simplification
 *
 *
 **********************************************************)

(*returns true if an effect does not contain any LockOp *)
let rec is_garbage (f:effect) : bool = 
 try
   List.iter 
   ( fun a -> 
      match a with 
      | LockOp _ -> 
        raise Not_found
      | Join el ->
        if List.exists 
          (fun x -> 
           let b = not (is_garbage x) in
          (* if b then 
           begin
            log "$$GARB_NODE :% s\n" 
                (effect2str_detail true true x);
           end;*) b
          ) el 
          then raise Not_found
      | NCall (ci,res) -> 
           let f' = must_resolved "is_garbage" res in
           if not (is_garbage f') then raise Not_found
      | RCall _ ->
           rs_not_implemented "is_garbage/RCall"
   ) f; 
   true
 with 
  | Not_found ->  false

let rec hasLockOps (f:effect) : bool =
try   
 List.iter (fun  a  ->
  (*log "Atomic effect : %s\n" (atomic_effect2str a);)*)
  match a with 
  | LockOp _ ->
    raise Not_found
  | Join el ->
    if List.exists hasLockOps el then raise Not_found
  | NCall ((_,_,vi),res) -> 
    let f' = must_resolved "hasLockOps" res in
    if hasLockOps f' then raise Not_found
  | RCall _ -> ()
 ) f ; false
with 
 | Not_found -> true



(* remove garbage nodes from effect.
 * *)
let sgarbage (f:effect) : effect =
let rec  sgarbage0 (f0:effect) : effect =
 List.fold_left (fun f' a  ->
  (*log "Atomic effect : %s\n" (atomic_effect2str a);)*)
  f' @
  match a with 
  | LockOp _ ->
    [a]
  | Join el ->
    let l' = List.filter (fun x -> not (is_garbage x)) el in
    begin match l' with
    | [] -> [] 
    | hd::[] -> sgarbage0 hd
    | hd::tl -> [Join (List.map sgarbage0 l')]
    end
  | NCall ((_,_,vi) as ci,res) -> 
    let f' = must_resolved "sgarbage0" res in
    let b = is_garbage f' in
    (*log "IS GARBAGE : %s  ? %b effect=%s\n" 
    (vi.vname) b (effect2str f');*)
    if b then []
    else [NCall (ci,mkresolved (sgarbage0 f'))]
  | RCall _ ->
    if hasLockOps f then
       rs_not_implemented "sgarbage0/RCall"
    else
       []
 ) [] f0

in sgarbage0 f

 (* compares two effects for equality 
 * this function assumes that effects
 * have already been simplified 
 * *)
let rec list_eq (eq:'a->'a->bool)
                (l1:'a list)
                ( l2:'a list): bool  = (*generic list equality*)
  match (l1,l2) with 
  | ([],[]) -> true
  | (hd1::tl1,hd2::tl2)  ->
    not (eq hd1 hd2) &&
    (list_eq eq tl1 tl2)
  | _ -> false


let rec
  aeffect_equal (af1:atomic_effect) (af2:atomic_effect):bool = 
  match (af1,af2) with
  |(LockOp ((loc1,_,_),b1,s1),LockOp ((loc2,_,_),b2,s2)) -> 
  (* b1 = b2 && (sexp_eq s1 s2) *)
     loc1 = loc2 (* cannot perform important optimizations !*)
  |(Join el1, Join el2) ->
    list_eq effect_equal el1 el2
  | (NCall ((cloc1,_,_),_),NCall ((cloc2,_,_),_)) ->
    cloc1 = cloc2 
  | (RCall ((cloc1,_,_),_),RCall ((cloc2,_,_),_)) ->
    cloc1 = cloc2 
  | _ -> false
and
   effect_equal (f1:effect) (f2:effect) : bool =
   list_eq aeffect_equal f1 f2 


(* shrink -- opposite of flatten*)
let rec shrink (f:effect) : effect = 

 let rec prefix (p:effect) (l:effect) : 
  (effect * effect * effect) = 
  match (p,l) with
  | ([],[]) -> ([],[],[])
  | (hd1::tl1,hd2::tl2) ->
    let b0 = aeffect_equal hd1 hd2 in
    if b0 then 
     let (a,b,c) = prefix tl1 tl2 in
     (hd1::a,b,c) 
    else 
     ([],p,l)
  | ([],_) -> ([],[],l)
  | (_,[]) -> ([],p,[])
 in

 let rec postfix  (p:effect) (l:effect) : 
    (effect * effect * effect) = 
  let (a,b,c) = prefix (List.rev p) (List.rev l)
  in  (List.rev a,List.rev b, List.rev c) 

 in
  
 List.fold_left 
 (fun (f0:effect) af1 ->
   f0 @
   match af1 with
   | LockOp a ->
     [LockOp a]
   | Join el0 ->
     let el = List.map shrink el0 in
     begin match el with
     | [] -> []
     | hd::[] ->  hd
     | hd1::hd2::tl ->

       let (pre0,hd1',hd2') = prefix hd1 hd2 in
       let (post0,_,_) = postfix hd1' hd2' in

       (*log "PREFIX OF :\n %s \nAND\n %s\n is\n %s\n\n\n" 
            (effect2str hd1) 
            (effect2str hd2)
            (effect2str pre0);*)

       let pp (p0,p1) f =
        let (pre0,_,b) = prefix p0 f in
        let (post0,_,_) = postfix p1 b in
         (pre0,post0) 
       in


       let (pre0',post0') = 
          List.fold_left (fun a f -> pp a f) 
                         (pre0,post0) tl in

       let pp'  f  =
        let (_,_,b)  = prefix pre0' f in
        let (_,_,c) = postfix post0' b in
        c
       in

       let el' = List.map pp' el in
       let b = List.for_all (fun f->f=[]) el' in
   
       if b then pre0' @ post0' 
       else pre0' @ [Join el'] @ post0' 
               

(*           
       (* define helper functions *)
       let post = split false false in
       let pre = split true true in
       let both f0 f1 = 
        let (a,_,_,d) = split false true f0 f1 in
        (a,d)
       in 
       let pp p0 p1 f = 
        let (p0',_,a,_) = pre p0 f in
        let (_,_,_,p1') = post p1 a in
        (p0',p1')
       in
       let pp' p0 p1 f = 
        let (p0',_,a,_) = pre p0 f in
        let (_,_,b,_) = post p1 a in
        log "\nBEFORE : %s\n\nNOW: %s\n\nA=%s\n\nP1=%s\n\nP0=%s\n\n\n\n" 
        (effect2str f) (effect2str b)
        (effect2str a) (effect2str p1)
        (effect2str p0);
        b
       in
       (* find the greatest common prefix
        * and greatest common postfix GCP*)
       let (p0',p1') =  List.fold_left
                        (fun (p0,p1) f -> pp p0 p1 f) 
                        (let (pa,pb) = both hd1 hd2 in
                           log "\n################\n%s\n################\n%s\n################\n"
                           (effect2str pa) (effect2str pb);
                           (pa,pb)) tl  
       in
          log "\n#######SS########\n%s\n################\n%s\n################\n"
            (effect2str p0') (effect2str p1');

       (* simplify paths using p0' and p1' *)
       let el' = List.map (pp' p0' p1') el in
       (* *)
       let all_empty = List.for_all (fun f->f = []) el' in

       if all_empty then p0' @ p1'
       else  p0' @ [Join el'] @ p1'  
*)       
    end
   | NCall (x,res) ->
     let f' = must_resolved "shrink/3" res in
     [NCall (x,mkresolved (shrink f'))]
   | RCall _ -> 
     (raise (Impossible "shrink/2") : effect)
 ) [] f 

 




(* simplify calls:
   *
   * removes NCall nodes and inlines their effects
   * except for the top-level ones
   * that are currently required as markers when translating
   * type-level effects to run-time efffects
   *
   * TODO: RCall  
 *)
let scall (f:effect) : effect =
let rec scall0 level (f:effect) : effect =
 List.fold_left
 (fun f' elt -> 
   f' @
   match elt with 
   | NCall (ci,res) -> 
    let f'' = must_resolved "scall0/NCall" res in
    if level = 0 then 
       (scall0 (level+1) f'') @ [NCall (ci,mkresolved [])]
    else  
       (scall0 (level+1) f'')
   | Join el ->
      ((Join (List.map (scall0 level) el))::[])
   | RCall _ ->
     raise (Not_implemented "scall0")
   | LockOp a ->
     [LockOp a] 
 ) [] f
in
scall0 0 f

(*************************************************************************** 
 *                      callVisitor class 
 *
 *
 *
 *
 ***************************************************************************)
module OrderedInt =
  struct
    type t = int
    let compare = compare
  end
module IntSet = Set.Make(OrderedInt)


module IntMap = Map.Make (
       struct
         type t = int
         let compare = Pervasives.compare
       end
      )
type effect_map_val = effect * 
                      effect option * 
                      int option
(************************************************************************)

(* Abstract class with utility methods
 * invokes callback for each lock instruction
 *)
(************************************************************************)
type stmtKind = 
    (* sid of Loop , break stmt ,total number of preds,
     * number of preds left *) 
 |   SWhile  of (int * stmt * int * int ref)
 |   SNormal of int ref (* number of preds left*) 


let txt = open_out "nodes.dot" 
let logtxt format =  Printf.ksprintf
                     (fun x ->Printf.fprintf txt "%s" x) format 

(*let cgtxt = open_out "cg.dot" 
let logcg format =  Printf.ksprintf
                   (fun x ->Printf.fprintf cgtxt "%s" x) format *)
 
let mklabel_s (s : stmt) : string =
    begin
      match s.skind with
        | If (_, _, _, loc)  -> 
              "if/" ^ (loc2str loc) 
        | Loop (_,loc,_,_) -> 
              "loop/" ^ (loc2str loc)
        | Goto (_,loc) -> 
               "goto/" ^ (loc2str loc)
        | Instr _ -> 
            "instr/" ^ (loc2str (Cil.get_stmtLoc s.skind))
        | Block b -> 
            "block/" ^ (loc2str (Cil.get_stmtLoc s.skind))
        | Return (_,loc) -> 
              "return/" ^ (loc2str loc) 
        | _ -> raise (Impossible "mklabel_s") 
    end 

   (* this type represents
    * (1)  the first one is the result of +/- in the effect
    * (2)  the second one is the initial # of - before the first +
    * (3)  the third one indicateds whether
    *      the first  + has been found
    * *)
type effect_count = int * int * bool 

let fnd_ec map (vid:int) = 
   begin match absFind IntMap.find vid !map  with
    | None ->   (0,0,false)
    | Some b -> b
   end 

let lock_vid = ref IntSet.empty
let addLockVid vid = 
   lock_vid := IntSet.add vid !lock_vid
let resetLockVid () =
   lock_vid := IntSet.empty
let iterLockVid f =
   IntSet.iter f !lock_vid

(******************************************************
 * WARNING:
 * resetLockVid SHOULD NOT be invoked within function
 * lock_counts
 **************************************************** *)
let rec lock_counts (f:effect) 
                  : effect_count IntMap.t ref = 
 let map  = ref (IntMap.empty) in
 let add s (vid:int) n = 
   (*let (n0,_,_) = n in
   log "Adding (%s) %s count =%d\n\n" s
   (varinfo2str (Cilinfos.getVarinfo vid)) n0;*)
    map := IntMap.add vid n !map
 in
 (* log "lock_counts effect : %s\n" (effect2str f);*)
 let ar (z:effect->unit) (f:atomic_effect) : unit =
  (*log "AR considering : %s\n" (atomic_effect2str f);*)
  begin  match f with 
(*  | Wildcard -> () *)
  | LockOp(_ ,is_acquire,sexp) -> 
    (*Crucial POINT! sexp info better be accurate*)
    let vid = (vid_of_sexp sexp) in
    let add' = add "LO" vid in
   (* log "\n\nChecking effect: %s\n" (atomic_effect2str f);*)
  (*  log "\nAdding vid=%d is_acquire=%b sexp=%s --- "
         (vid) (is_acquire) (sexp2str sexp);*)
    addLockVid vid; (*  fast lookup for all locks*)
    begin match fnd_ec map vid with
     | (n0,n1,false) ->
      (* log "false n0=%d  n1 = %d is_acquire=%b vid=%d e=%s\n" 
            n0 n1 is_acquire vid (sexp2str sexp);*)
       if is_acquire then
        add' (n0+1,n1,true)
       else 
        add' (n0,n1+1,false)

     | (n0,n1,true) ->
      (* log "true n0=%d  n1 = %d is_acquire=%b vid=%d e=%s\n" 
           n0 n1 is_acquire vid (sexp2str sexp);*)
       if (not is_acquire) && (n0-1 < 0) then
         err "Found negative effect (lock_counts) for %s\n\n" 
             (sexp2str sexp)
       else 
         add' (n0+(if is_acquire then 1 else (-1)),n1,true)
    end     
  | Join el -> 
   let doAdd key a b = 
    if a <> 0 || b <> 0 then
    begin
     let (e,f,_) = fnd_ec map key in
     if e < b then
      add "IF1" key (a,b-e+f,true)
     else
      add "IF2" key (a+e-b,f,true)
    end 
   in
    ignore(arePathsEqualCb effect2str el doAdd)

(*  | Branch (f1,f2) -> 
     (* obtain the effects for the two branches*)
     let map0 = lock_counts f1 in 
     let map1 = lock_counts f2 in
     let doLock key = 
      let (a,b,_) = fnd_ec map0 key in
      let (c,d,_) = fnd_ec map1 key in
      (* perform check: the effects of the two branches
       * must match *)
       if (a <> c) || (b <> d) then
          err "The effect sum of the two branches differs.";
      (* update lock counts *)
      if a <> 0 || b <> 0 then
      begin
       let (e,f,_) = fnd_ec map key in
       if e < b then
        add key (a,b-e+f,true)
       else
        add key (a+e-b,f,true)
      end
     in
     (*compare the effects of the two maps and
      * add the resulting effects to the current map*)
     iterLockVid doLock*)

  | NCall(_,f_res) -> 
        z (must_resolved "lock_counts/NCall" f_res)
  | RCall (_,f_opt_ref) -> 
    begin match !f_opt_ref with
     | Some f' -> z f' 
     | None -> 
        raise (Impossible "lock_counts/RCall: unresolved")
    end
        

(*  | Call_resolved (_,_,f)  
  | RecursiveCall_resolved f -> z f 
  (*| Iteration_resolved f  ->  z f*)
(*  location of call, the call itself, location of fun being called,*)
  | Call_unresolved _ -> 
        rs "lock_counts/Call_unresolved"
  | RecursiveCall_unresolved _ -> 
        rs "lock_counts/RecursiveCall_unresolved"
(*  | Iteration_unresolved _ -> 
        rs "IsZeroEffect/Iteration_unresolved"*)*)
  end in 
 let rec fr (f:effect) : unit = List.iter (ar fr) f  in
  fr f; (* collect counts *)
  map   (* return map *)

and
 arePathsEqualCb
 (fstr : effect -> string) 
 (el : effect list)
 (upd: int->int->int->unit) : bool =  
 let update_parent m = 
    iterLockVid  
     (fun vid -> 
       let (n0,n1,_) = fnd_ec m vid in
       upd vid n0 n1
     )
 in
 match el with 
  | [] -> true 
  | only::[] -> 
     update_parent (lock_counts only); 
     true (* is this correct ?*)
  | hd::tl  ->
    let map0 = lock_counts hd in
    let ok = ref true in
    let doCheck f map1 key = 
     let (n0',n1',_) = fnd_ec map0 key in
     let (n2,n3,_) = fnd_ec map1 key in
     let zero = n3 = 0 && n2 = 0 in

     let (n0,n1) =
       if (n0' = 0 && n1' = 0) && (not zero)  then
       begin
         (*use this tuple for subsequent comparisons *)
          map0 := IntMap.add key (n2,n3,true) !map0;
         (n2,n3)
       end
        else (n0',n1')  
     in

     if (zero && n0 <> n1) || 
        ((not zero) && (n0 <> n2 || n1 <> n3)) then
     begin
        ok := false;
        err "paths are not equal. \
             Example:\nPath 1: %s\nPath 2: %s\nVar:%s\n\tPath 1: consumes (-) = %d \
             produces (+) = %d\n\tPath 2: \
             consumes (-) = %d produces (+) = %d\n" 
             (fstr hd)
             (fstr f)
             (varinfo2str (Cilinfos.getVarinfo key))
             n1 n0 n3 n2;
             exit (-1);
     end
    in
    (* iterate through and update map0 as necessary *)
    List.iter (fun f -> iterLockVid
                        (doCheck f (lock_counts f))) tl;
    (* update effect counts using the first path *)
    update_parent map0;
    !ok


let isLoopEffectPreserved
   (inp:effect) (loop:effect):bool=
 let map0 = lock_counts inp in 
 let map1 =(* (log "\n\n");*) lock_counts loop in
 let ok = (* (log "\n----zzzzz\n");*) ref true in
 let doCheck a (n0,n1,_) =  
  let (n2,_,_) = fnd_ec map0 a in
    (*log "vid=%d n0=%d n1=%d n2=%d \n" a n0 n1 n2;*)
    if (n0 <> n2) || (n1 > n2) then 
    begin
       ok := false;
    end 
 in
 IntMap.iter doCheck !map1;!ok 

let arePathsEqual
 (fstr : effect -> string) 
 (el : effect list) : bool =  
   (* log "Are paths equal : %s\n" 
    (effect2str_detail true true [Join el]); *)
    let b = arePathsEqualCb fstr el (fun _ _ _ -> ()) in
   (* log "%b\n\n" b;*)
    b

(* simple structural equality*)
(*let effects_equal (f0:effect) (f1:effect) =*)


let branch_of_paths (el: effect list) : effect =
(*TODO: perform simplifications before creating the
 * branch: 
 *    0) remove duplicate paths
 * *)
 match el with
  |  [] -> [] 
  |  hd::[] -> hd
  |  _ -> (Join el)::[]
(*
 begin match el with
 | [] -> [] 
 | hd::[] -> hd
 | hd1::hd2::tl ->
   let fl = (fun br f -> Branch(br::[],f)) in
   (List.fold_left fl (Branch(hd2,hd1)) tl)::[]
 end 
*)

let lock_count_map = ref (IntMap.empty :  bool IntMap.t)
let will_backtrack vid = 
 begin  try
     IntMap.find vid !lock_count_map 
  with Not_found ->
    rs_impossible "will_backtrack" 
  end

let calc_counts vid f = 
 let map = lock_counts f in 
 let is_empty =
  try
   IntMap.iter (fun  _ (n,_,_) -> 
               if n > 0 then raise Not_found) !map ;
   false               
  with
   Not_found -> true
 in
 log "Will backtrack : %s ?  %b\n" 
     (vid2str vid) is_empty;
 lock_count_map := IntMap.add vid is_empty !lock_count_map 

(*******************************************************)
class callVisitor = object (self) inherit nopCilVisitor

  val mutable context_fundec=  dummyFunDec
  val mutable context_loc   =  Cil.locUnknown 
  val mutable call_loc      =  Cil.locUnknown 
  val mutable call_stmt     =  Cil.dummyStmt
  val mutable fun_e         =  (SizeOfStr "")
  val mutable args_e        =  []
  val mutable fun_loc       =  Cil.locUnknown
  val mutable fun_fundec    =  Cil.dummyFunDec 
  val mutable fun_name      =  ""
  val mutable cur_instr     =  Cil.dummyInstr 
  val mutable bad_glob_cnt  =  0
  val mutable fun_varinfo   = Cil.dummyFunDec.svar

  val lval_map =  Hashtbl.create 117

(***********************)

method getMods (():unit) : sum_typ =
 let hfold key value tl = value::tl in
 Hashtbl.fold hfold lval_map [] 

method addSum_aLval lv =
try
(*
let scope2str (s : Scope.scope) = 
   match s with 
    SFormal i  -> Printf.sprintf "SFormal(%d)" i
  | SGlobal -> "SGlobal"
  | SFunc -> "SFunc" 
  | STBD  -> "STBD" in*)
let scopeOf lv = Scope.decideScopeVar context_fundec 
                 (Lvals.findBaseVarinfoLval  lv
                   (* (Lvals.abs_of_lval lv)*)
                 ) in
let s = scopeOf lv in
(* log "Adding lv=%s with scope=%s\n"
 (alval2str lv) (scope2str s);*)
 Hashtbl.replace lval_map lv  (lv,s) 
with _ -> log "Could not get scope for : %s\n"
                         (alval2str  lv)

method private doCall e =
(* let printLVal lv =
  let (_,lvist) = SPTR.derefLvalAtInstr call_stmt cur_instr lv 
  in begin
   Printf.printf "DEREF  of expression : %s is\n %s\n\n"
   (exp2str e)
   (abs_list2string "," 
   (fun x-> any2str Lvals.d_alval x) 
   lvist)
   end
 in*)
let setF f = 
   fun_varinfo <- f;
   fun_loc <- f.vdecl; 
   fun_name <- f.vname 
in
match e with
 | Lval( (Var(f),_) (*as lv*) ) -> 
   setF f;
  (*printLVal lv;*)
  begin match f.vname  with
   | "mypthread_mutex_unlock" -> self#doRelease () 
   | "mypthread_mutex_lock" ->   self#doAcquire ()
(* | "mypthread_cond_signal"    ->   self#doSignal ()
   | "mypthread_cond_broadcast" ->   self#doSignal ()
   | "mypthread_cond_wait"  ->   self#doWait ()*)
   | _ ->
         self#doNormal () 
  end
 (* Indirect call *)
 | Lval( (Mem(Lval(Var(f),_)),_) (*as lv*)) -> 
       setF f;
 (*printLVal lv;*)
       self#doIndirect ()
 | _ ->   
       self#doIndirect ()

method private doInstr  (i:instr) = 
cur_instr <- i;
begin match i with
  Call(_,e,el,loc) -> call_loc <- loc;
                      fun_e <- e;
                      args_e <- el;
                      self#doCall e
| Set (lv,e,loc) -> 
 if loc = locUnknown then () 
 else let chk_typ t' =
 begin  match t' with 
    TNamed (ti,_) ->  
      if ti.tname =  "pthread_mutex_t" then ExistsTrue
      else ExistsMaybe
  | TComp (comp, a) -> 
      let chk f =
         (*log "Checking field : %s\n" f.fname; *)
         f.fname = "__data" (*FIXME: ugly hack.unsound!*)
      in if  List.exists chk comp.cfields  then
         ExistsTrue
      else 
         ExistsFalse
  | TPtr _ -> ExistsFalse
  | _ ->  ExistsMaybe
 end in
 let lvtyp =  (Cil.typeOfLval lv) in 
(* log "@@@@@ Location : %s assignment lv typ = %s\n"
(loc2str loc )  (any2str d_plaintype lvtyp);*)
 if chk_typ lvtyp = ExistsTrue then 
    err " Location : %s => Attempted to assign pthread_mutex_t in %s\n" 
      (loc2str loc)
      (any2str d_instr  i) ;


(*log "Is Shareable : lv=%s ? %b\n" 
 (any2str d_lval lv) 
 (opt2bool(Shared.isShareable context_fundec lv))*)


 (*(opt2bool (Shared.isShareableAbs context_fundec ((Lvals.abs_of_lval
 lv))))*)
 ;

(* save summary information --- unsound if 
 * FIXME: unsound if proper control-flow is added *) 
let mustPt, targets = 
     (SPTR.derefLvalAtInstr call_stmt i) lv in
(*
let decideScopeVar curFunc vi : scope =
 * *)
List.iter (self#addSum_aLval) targets

(* | Set (lv,e,loc) -> 
   if fun_fundec.svar.vname  = "BAD_MUTEX_FN" then begin
      if bad_glob_cnt = 0 then 
         good_init_e <- e 
      else  
         Hashtbl.add bad_globs bad_glob_cnt (good_init_e,e)
     ;bad_glob_cnt <- bad_glob_cnt + 1  
   end else  begin
      Printf.printf "SET : %s is zero assignment  ? %b\n"
      (exp2str (Lval lv)) (isZero e);
      val typeSig: typ -> typsig

   end*)
 | _               -> ()
end


(*method vinst  i =
 match  self#doInstr i with
   Some il -> ChangeTo il
 | None    -> DoChildren*)

method vstmt s =
 match s.skind with 
  Instr il ->
     call_stmt<-s;
     List.iter (fun (i :instr)->self#doInstr i) il;   
     DoChildren
 | _   -> 
     DoChildren

method vglob g = match g with
  GFun (fdec,loc)   -> context_fundec <- fdec ; 
                       context_loc <- loc;
                       DoChildren
 | _                -> SkipChildren 

(**********************************************************
 *  
 *                      new stuff
 *
 *
 *********************************************************)
val mutable main_vid = (-1) 
val mutable instr_effect = ([] : effect)

(* all RCall will point to the same rec_effect*)
val mutable rec_effect = (ref None : effect option ref)

(* Global scope: a map from
 * ===> vid -> (fundec,effect)
 * ===> the following data structures should NEVER be reset
 * *)
val mutable effect_map = IntMap.empty

(* Local scope: fundec specific
 * The following data structures should be reset
 * each time a new fundec is visited
 * *)
val worklist = Queue.create ()
val mutable visited  = IntSet.empty
val mutable leafs  = IntSet.empty

(* map : continue labels -> while stmt sid*)
val mutable cont_sid  = IntMap.empty
val mutable sid2effect = IntMap.empty 
(*val mutable count = 0 *)
val mutable edge_count = 0; 

(************ new methods **************************)
method thisEffect () : effect = instr_effect
method resetThisEffect () : unit =  instr_effect <- [] 

method addEffect f =  
 instr_effect <-  instr_effect @ [f] 

method private debug_lockop (arg:Cil.exp)
                            (s:sexp)
                            (must:bool) : unit  =
  log "Location: %s ---> Alias of expression : %s is\n %s (%s) Sharable ? : %s\n\n"
  (loc2str call_loc)
  (exp2str  arg) 
  (if must then "must" else "may")
  (sexp2str s ^ ":::" ^ (detail_sexp s))
  ( match arg with
     Lval lv -> 
      if opt2bool(Shared.isShareable context_fundec lv) then
      "true" else "false" 
   | _ -> "dunno" 
  )

method private getSexp arg = 
 let (must,elist) = 
       SPTR.getAliasesAtInstr call_stmt cur_instr arg in
 let is_singleton = match elist with _::[]->true|_->false in
  if not must || not is_singleton then begin
     err "Found a may expression (%s) at location %s\n"
     (exp2str arg) (loc2str call_loc);
     None 
  end else begin
   let  cs = (List.hd elist) in
   (*log "AEXP: \n%s\n" (aexp2str cs);*)
   let  s  = simplify_aExp 
             (fun _->context_fundec)  (*FIXME*)
             cs 
   in
      Some s
  end
  
method doLockOpArg arg is_acquire  = 
 let calli = (call_loc,fun_e,arg) in  
  match self#getSexp arg with
   | Some s ->
       self#addEffect (LockOp(calli,is_acquire,s))
      (*;log "\nLOCK OP : %s\n" (stmt2str call_stmt)*)
      (*;log "\nLval!!! : %s\n" (alval2str (CMem cs,NoOffset))*)
      (*;self#debug_lockop arg s must*)
   | _ -> 
       ()

method doLockOp is_acquire = 
 match args_e with 
  | arg1::[] -> 
    ( self#doLockOpArg arg1 is_acquire : unit)
  | _ ->
    err "Invalid args at %s\n" (loc2str call_loc)

method doSignal () =
(* effect of signal: c+ m- *)
 self#doLockOp true; 
 self#doLockOp false

method doWait () =
  match args_e with 
   | arg1::arg2::[] ->
     let cond_lk = self#doLockOpArg arg1 in 
     let mutex_lk = self#doLockOpArg arg2 in 
     (* effect of cond wait: c+ m- c- m+ *)
     cond_lk true;
     mutex_lk false;
     cond_lk false;
     mutex_lk true
   | _ ->
     err "Invalid args at %s\n" (loc2str call_loc)


method simpleEffectOf vid = 
 let (_,f,_) = self#effectOf vid in f

method effectOf (vid:int):(Cil.fundec*effect*Cil.varinfo list) = 
 begin match absFind IntMap.find vid effect_map  with
 | Some x -> 
   x
 | None ->
   (* some reason glob_init is never visited.
    * So this hack fixes things
    * *) 
   let fd = getGlobInit (opt2some !cilfile)  in
   if fd.svar.vid = vid then
      (fd,[],[])
   else
   begin
     let fd_opt = (vid2fundec_opt vid) in
     (* if the function descriptor exists 
      * then there's a bug 
      * *)
     if opt2bool fd_opt  then
      let str = (opt2some fd_opt).svar.vname in
      rs_impossible
      "effectOfCalledFun:\
         Bottom-up issue? or mutually recursive function => %s " str
     else
      (dummyFunDec,[],(dummyFunDec.svar)::[])
    (*  let globs= (opt2some !cilfile).globals in
    log "GLOBS : %d\n" (List.length globs); 
     List.iter
     (fun g ->
        match  g with 
         |   GVarDecl (vi,_) ->
           log "VARDECl : %s\n" vi.vname
         |  _ -> () 
      ) globs ;*)
     (* if str = "mypthread_mutex_lock" ||
         str = "mypthread_mutex_trylock" then
         (dummyFunDec,[],(dummyFunDec.svar)::[])
      else*)
      (* *)
   end
 end

method vidOfMain() : int = 
   if main_vid = (-1) then 
      raise (Impossible "vidOfMain"); 
   main_vid

method effectOfMain () : effect =
   if main_vid = (-1) then 
      raise (Impossible "effectOfMain"); 
   let (_,f,_) = self#effectOf main_vid in f

(**)
method private makeCallEffect vid =  
  let aliases x= snd (SPTR.getAliasesAtInstr call_stmt cur_instr x)in 
  let simplify=(simplify_aExp (fun _-> context_fundec)) in
  let simparg arg=
   (*  let s1 =  (exp2str arg) in*)

         List.map 
         (fun s -> 
          (*let s2 = (aexp2str s) in
           log "ARG = %s alias = %s\n" s1 s2;*)
           simplify s
         )
         (aliases arg) 
  in
  let (args0:sexp list list)=List.map simparg args_e in

  (* obtain the effect of*)
  let (fd,f_fun,formals) = self#effectOf vid in
  let no_args = List.length args0 in
  let no_formals = List.length formals in
  let fname =  (fundec2str fd) in
(*  log "Function=%s location=%s args0=%d formals=%d\n" 
      fname (loc2str fun_loc)
      (no_args) (no_formals);*)
 if fd.svar.vid = dummyFunDec.svar.vid then []
 else 
 begin
  let is_vararg  = (*FIXME:varg fun ptrs ?*) 
     match typeOf fun_e with 
     | TFun(_,_,b,_) -> b 
     | _ -> false 
  in
  let same_args = (no_args  = no_formals) in

  if not (same_args || is_vararg) then
   err "Function call \"%s\" at location %s: #formal_args <> #actual_args (formals=%d, actuals=%d).Exact statement: \"%s\""
        fname (loc2str call_loc) (no_formals) (no_args) (stmt2str call_stmt);

  let args0' = if same_args then args0 
               else
                snd(List.fold_left 
                   (fun (i,lst) e -> (i+1,
                      if i < no_formals then lst @ [e]
                      else lst)) (0,[]) args0)
  in
  let comb = List.combine formals args0'  in   

(*let rtime = rstart_new "doCALL BEFORE!!" in
print_mb "doCALL BEFORE" ;
let pr s =
      rend_time ("##### AFTER doCALL!! " ^ s) rtime;
      print_mb  ("#### AFTER doCALL!! " ^ s)
in*)

  (*IMPORTANT: remove garbage effects
   *  - DO NOT REMOVE sgarbage as memory usage will
   *    dramatically increase 
   * *)
  let f_fun_0 = sgarbage f_fun in 
  let subst = substEffect call_loc comb f_fun_0 in
  (*FIXME: filterEffect attempts to filter all 
   *       operations on locks that are created
   *       within f_fun.
   *       It doesn't do everything correctly.
   **)
  let f_fun' = filterEffect call_loc subst in
(*  let e2s = effect2str in 
  log "Original effect:\n%s\nAfter subst\n:%s\
       \nAfter filtering:\n%s\n"
   (e2s f_fun) (e2s subst) (e2s f_fun');*)
(*  pr ("@@@@@@ 0  @@@@@@" ^ fun_name ^ " %%%% ");*)
  let f_fun''= 
     try 
        let b1 = hasErrors() in
        let f_out = morePreciseEffect aliases f_fun' in
        let b2 = hasErrors () in
        if (not b1) && b2 then 
           err "Function: %s\n" fd.svar.vname;
        f_out
     with Impossible _ -> f_fun'
  in

(*  pr ("@@@@@@ 1  @@@@@@" ^ fun_name ^" %%%% ");*)
  f_fun''
 end

(* Dummy methods that can be overridden*)
method doNormal   () =  
 let ci = (call_loc,fun_loc,fun_varinfo) in 
 if Cil.compareLoc context_loc fun_loc = 0 then
   self#addEffect  (RCall(ci,rec_effect))
 else (*if not recursive call *)
 begin
  let f = self#makeCallEffect fun_varinfo.vid in
  if f  <>  []  then
    self#addEffect (NCall(ci,mkresolved f))
 end


(*val indirect_call_map = Hashtbl.create 50*)

method doIndirect () =
let aliases = (alias_map call_loc) in
let vids_opt  = List.map varinfo_of_name aliases in
let (el:effect list) = List.fold_left
           (fun lst x ->
              match x with
              | Some vi ->
                let (t1,t2)=  (typeOf fun_e,vi.vtype) in
                let (ts1,ts2) = (typeSig t1,typeSig t2) in
                if Util.equals ts1 ts2 then
                begin
(*                 log "INDIRECT: typeof fun_e = %s  type of alias = name = %s and type %s\n" 
                     (typ2str (typeOf fun_e))  vi.vname (typ2str vi.vtype);*)
                 (self#makeCallEffect vi.vid)::lst
                end
                else lst
              | None -> lst
           ) ([]:effect list)  vids_opt 
in 
(*Hashtbl.add indirect_call_map call_loc *)
let f0 =  match el with 
           | [] -> [] 
           | hd::[] ->  hd
           | _ ->
             if List.for_all (fun x -> x = []) el then
              []
            else
             [Join el] 
in           
if f0 <> [] then
begin
   let ci = (call_loc,locUnknown,fun_varinfo) in 
   let f = NCall (ci,mkresolved f0) in
    self#addEffect f
    (*;log "INDIRECT: %s effect : %s\n"
      (abs_list2string "," (fun x->x) aliases)
      (effect2str_detail true true [f])*)
end      
      


method doAcquire  () = self#doLockOp true
method doRelease  () = self#doLockOp false

method setInputEffect 
 (from_s:stmt) (effect:effect) (to_s:stmt) : unit =

    logtxt "\"%d\" -> \"%d\" [label=\"%s/%d\"];\n"
    (from_s.sid) (to_s.sid) 
    (let f = effect2str effect in
     if f = "" then "empty" else f )
    edge_count
    ;
    edge_count <- edge_count + 1; 
(*log "\nINPUT_FLOW FROM\n\n\n  %s \n\n\nTO(sid=%d preds=%d)\n\n\n%s\n\n\n\nWITH EFFECT : %s\nEND_INPUT_FLOW\n" 
    (stmt2str from_s) 
    (to_s.sid) 
    (List.length to_s.preds) (stmt2str to_s) 
    (effect2str effect); *)
let add b = sid2effect <- IntMap.add to_s.sid b sid2effect in
let get_cont () : (int*stmt*int*(int ref)) option =absFind IntMap.find to_s.sid cont_sid in

match absFind IntMap.find to_s.sid sid2effect with
| Some (f_in::tl,None,
         (SWhile (while_sid,break_stmt,total,preds))) -> (* cont node *)
  let pf = effect in (*postfix f_in effect in*)
  (*log "\nLOOP_EFFECT = %s\nENV+LOOP_EFFECT = %s\n" 
  (effect2str pf) 
  (effect2str effect);*)

  preds := !preds - 1;
  (*log "\nENV+LOOP_EFFECT (preds=%d) = %s\n"  
      (!preds) (effect2str pf);*)

  if not (isLoopEffectPreserved f_in effect) then
    err "Iteration effect sum <> 0:\n-----------------\nEffect:%s
    \n-----------------\nflows from \
    statement\n-----------------\n%s\n-----------------\nto \
    statement\n-----------------\n%s\n-----------------\n"
    (effect2str pf)
    (stmt2str from_s) (stmt2str to_s)
  (* else*)
  ;
  begin
     (* propagate flow to break stmt.
      * The preds of break SHOULD not be decremented
      * *)
     add ((f_in::tl)@[effect],None,
          (SWhile (while_sid,break_stmt,total,preds)))
     (*log "HERE!!@!\n";
     self#setInputEffect to_s effect break_stmt*)
  end     

   (* there's nothing to add in this case *) 
| Some (f',None,SNormal preds) ->  (* normal node *)
      preds := !preds - 1;
      (* log "\nNORMAL_NODE1: %s\n" 
      (abs_list2string "|||" (effect2str ) f');*)
      add (effect::f',None,SNormal preds)
| None -> 
    (*  log "\nSET_INPUT_EFFECT : NEW NODE ->";*)
      let sk = begin match get_cont () with
       | Some (a,b,total,preds) ->
             (*log "SWHILE = %d \n" !preds;*)
             if !preds > 0 then preds := !preds -1;
             SWhile (a,b,total,preds)
       | None ->
         (*log "SNormal \n";*)
         let preds = List.length to_s.preds in
         let preds' = if preds > 0 then preds -1 else 0 in
         SNormal (ref preds')
      end  in
      add ([effect],None,sk)

(*| Some (f',f_opt ,SNormal preds) ->  (* normal node *)
      preds := !preds - 1;
      let f'' =  (Branch (f',effect))::[] in
      log "\nNORMAL_NODE2: %s\n" (effect2str f');
      add (f'',Some f'',SNormal preds)*)

| Some (_,(b:effect option),(c:stmtKind)) ->
   
      raise 
   (Impossible (Printf.sprintf 
    "setInputEffect %b to_s.sid=%d to_s=%s\nfrom.sid=%d =>from_s=%s\n" 
     (opt2bool b) to_s.sid 
     (stmt2str to_s) from_s.sid (stmt2str from_s)
   )) 

method setOutputEffect   
 (s:stmt) (tail :effect) : effect  =
let add b = sid2effect <- IntMap.add s.sid b sid2effect in
match absFind IntMap.find s.sid sid2effect with
 | Some (f'',None,sknd) ->
  let is_norm = 
   begin match sknd with 
    | SNormal preds->
        if !preds > 0 then 
           raise (Impossible  (Printf.sprintf
                     "setOutputEffect: SNormal: preds=%d stmt=%s"
                     !preds (stmt2str s)
           )); true
    | SWhile (_,_,_,preds) -> 
        if !preds > 0 then 
           raise (Impossible 
                  (Printf.sprintf
                     "setOutputEffect: SWhile: preds=%d"
                     !preds
           ));false
          (*() *)
   end 
   in
   let f0 = if is_norm then f'' else
            begin match f'' with
             | _::tl -> tl
             | [] -> [] 
            end 
   in
   let f' = List.map (fun f->
                      let f' = shrink (sgarbage f) in
                      (*log "CONVERTING:\n%s\n\nTO\n\n%s\n\n"                        (effect2str f) 
                        (effect2str f');*)
                        f'
                     ) f0 in
   if not (arePathsEqual 
            (effect2str_detail true true) f') then
      err "Branches are not equal"; 
   let f'' = (branch_of_paths f') @ tail in
   add (f',Some f'',sknd);
   (*log "\nOUTFLOW = %s (is_normal=%b)\n\n"
   (effect2str f'') (is_norm); *)
    f''
 | Some (_,Some _,_) ->
    raise (Impossible 
      (Printf.sprintf "setOutputEffect@@@: %s\n"
       (stmt2str s)
     ))
 | None -> 
    raise (Impossible 
      (Printf.sprintf "setOutputEffect$$$: %s\nPREDS:%s\nSUCCS:%s"
       (stmt2str s)
       (abs_list2string "\n----\n" stmt2str s.preds)
       (abs_list2string "\n----\n" stmt2str s.succs)
     ))

method doStmt (stmt0:Cil.stmt) : effect =
 let enq e = Queue.add e worklist in 
 let deq () = Queue.take worklist in
 let notEmpty () = not (Queue.is_empty worklist) in
 let already_visited stmt0 = IntSet.mem stmt0.sid visited in
 let mark_as_visited stmt0 = visited <- IntSet.add stmt0.sid visited in
 let addLeaf stmt0 = 
    (*log "Adding leaf (sid=%d): %s\n\n" (stmt0.sid) (stmt2str stmt0) ;*)
     begin match absFind IntMap.find stmt0.sid sid2effect with
     | Some (_,_,SWhile _) -> 
            (*log "LEAF is WHILE\n"*) () 
     | Some (_,_,SNormal preds) -> 
           (* log "LEAF PREDS are %d (real preds=%d)\n" 
                (!preds) (List.length stmt0.preds);*)
            (*if !preds > 0 then 
               List.iter 
               (fun s-> 
                 log "LPRED (sid=%d): %s \n" s.sid  (stmt2str s) ;
                 List.iter 
                   (fun s-> log "\tSC (sid=%d): %s\n"
                            s.sid (stmt2str s)) s.succs 
               )
               stmt0.preds*) ()
     | None -> ()
          (*  log "LEAF NONE is %d\n"(List.length stmt0.preds) *)
          (* (List.length s.preds) < 2*)
     end;
    leafs <- IntSet.add stmt0.sid leafs in
 let addCont s1 s2 s3 = 
    let preds = (List.length s1.preds) in
    cont_sid <- IntMap.add s1.sid 
                        (s2.sid,s3,preds,ref preds) cont_sid in

 let pred s =
    begin match absFind IntMap.find s.sid sid2effect with
    |  Some (_,_,SWhile (_,_,_,p) ) -> !p
    |  Some (_,_,SNormal p) -> !p
    |  None -> -1
 end in

 let set_succs s f' = 
   let succ_inp s' =
(*     log "\nSET_SUCC\n\nSUCC (succ_sid=%d):\n%s" s'.sid (stmt2str s');*)
      self#setInputEffect s f' s' in
(*   log "\nBEGIN_SUCC\n"; *)
   List.iter succ_inp s.succs;
(*   log "\nEND_SUCC\n------------------------------\n" *)
 in

 let set_output b s f = 
  (*  log "\n------------------------------\nUPDATE STMT (sid=%d actual preds=%d remaining=%d):%s\n"
     (s.sid)    (List.length s.preds)  (pred s)
    (stmt2str s) ;*)

    let f' = self#setOutputEffect s f in
    begin match b with 
    |  Some break_stmt ->
      (*finalize break_stmt *)        
      self#setInputEffect s f' break_stmt;
      if (pred break_stmt) =  0 then
      begin
         let f'' = self#setOutputEffect break_stmt [] in
         set_succs break_stmt f'';
         List.iter enq break_stmt.succs; (* push successors *)
         mark_as_visited break_stmt; (* termination *) 
      end
   | None ->
       set_succs s f'
   end
 in

 let update_state s f =
  let print_skip id : unit =
(*   log "\n\n\nSKIPPING[%d] (sid=%d) (real preds=%d) (remaining=%d) :
          %s\nPREDS:\n%s\n\n\n\n" 
    id
    (s.sid)
    (List.length s.preds) (pred s)(stmt2str s) 
    (abs_list2string "\nPRED:" stmt2str s.preds)*) ()
  in

  let doit () =
   begin
       (*log "\nDOIT (preds=%d sid=%d): stmt = %s\n" 
       (List.length s.preds)
       (s.sid)
       (stmt2str s); *)
       List.iter enq s.succs; (* push successors *)
       mark_as_visited s; (* termination *) 
       set_output None s f 
   end
  in
  if (List.length s.succs) = 0 then addLeaf s; (*gather leafs*)
   (* all preds have updated input flow*)
  begin match absFind IntMap.find s.sid sid2effect with
     | Some (inp,_,
            SWhile (_,break_stmt,total,preds)) -> (*used to be true*)
      (*  log "BEFORE UPD FLOW: SWHILE: total=%d preds=%d\n"
         (total) !preds;*)
        if !preds = (total-1)  then
        begin
           List.iter 
           (fun s'-> (*log "\nsid=%d PUSHING sid'=%d\n"
           s.sid s'.sid;*) enq s' ) s.succs; (* push successors *)
           (*set input of succs*)
           match inp with
            | hd::[] ->
              set_succs s hd
            | _ ->
              raise 
              (Impossible "doit/Some/SWhile")
        end;

        if !preds = 0 then
        begin
         mark_as_visited s; (* termination *) 
         set_output (Some break_stmt) s f (* propagate output*) 
        end
        else print_skip 0 
     | Some (_,_,SNormal preds) ->
         if !preds = 0  then doit () 
         else print_skip 1
     | None -> 
       if (List.length s.preds) < 2 then doit () 
       else print_skip 2
  end 
 in
 let initial_state ()= 
    enq stmt0;  (* push statement *)
    self#setInputEffect stmt0 [] stmt0 in

 logtxt "digraph G {\n"; 
(* logcg "digraphG{\n";*)

 initial_state (); 
 (*log "INITIAL STATE DONE\n";*)
 while notEmpty () do
 let stmt = deq () in
   
  if already_visited stmt then begin
  (* log "\nALREADY VISITED %s \nEND VISITED\n" 
       (stmt2str stmt)*) ()
  end else begin

     logtxt  "\"%d\" [shape=\"box\",label=\"%s#%d\"];\n" 
     (stmt.sid) (mklabel_s stmt) (stmt.sid);


   (* let cnt = count in (* give a uid to the stmt *)
    count <- count + 1;  (* increment the count*)*)

(*    log "STMT (%d) : %s #succs=%d #preds=%d\n" 
            cnt (stmt2str stmt) (List.length stmt.succs)
            (List.length stmt.preds); (* debugging *)*)

    self#resetThisEffect ();
    begin match stmt.skind with 
     | Block b ->
       ()
     | Instr il -> 
       call_stmt<-stmt;
       List.iter (self#doInstr) il
            (*  log "Instruction (%d) : " (List.length il) ;
            List.iter (fun i ->
            log "Istr %s\nEND INSTR\n " 
            (any2str d_instr i)) il*)
     | Return (e_opt,loc) ->
       () (*no successors!*) 
       (*there are no side-effects in e_opt by defn of CIL*)
     | If (exp,block0,block1,loc) ->
           (*stmt.succs <- List.rev stmt.succs ;*)
       (*  log "IF ... block1 %d\n\n" 
         (List.length block1.bstmts) ;*)
       ()
       (* log "If stmt\n";*)
     | Loop (block,loc,cont_stmt_opt,break_stmt_opt) ->
       begin match (cont_stmt_opt,break_stmt_opt) with
        | (Some cont_stmt,Some break_stmt) ->
           (*letprn_s msg s  = log "\n%s\n%s\n" msg (stmt2str s) in*)
          (* log "\nLoop!!\nCONTINUE_LABEL (sid=%d) preds %d break_stmt:%s\n" 
               (cont_stmt.sid)
               ((List.length cont_stmt.preds))
               (stmt2str break_stmt)
               ;
            List.iter (fun s-> log "\nPRED: %s\n"
                         (stmt2str s)) cont_stmt.preds;*)

            addCont cont_stmt stmt break_stmt;
            (*ENSURE that break does only prop effects
             * once cont is computed
             * *)
            break_stmt.preds <- cont_stmt::break_stmt.preds
        
        | _ ->
           raise (Impossible "Loop case. CFG not computed ?")  
       end       
     | Goto (tgt_stmt_ref,loc)  -> 
     (*  log "GOTO: sid=%d tgt_sid=%d\n" 
       ( stmt.sid) ((!tgt_stmt_ref).sid);*)
       ()
     (* the following two are MVSC stuff ... not supported*)     
     | TryFinally _ ->  
       raise (Not_implemented "TryFinally")
     | TryExcept  _  -> 
       raise (Not_implemented "TryExcept")
     (* these are simplified away so
      * there's no need to implement them*)
     | Switch(e,block,stmt_list,loc) -> 
       raise (Impossible "Switch")
     | Break(loc) -> 
       raise (Impossible "Break")
     | Continue(loc) ->
       raise (Impossible "Continue")
    end ; (*end statement matching *)

(*    rend_time "##### LOOP_before_upd" rtime;*)
    update_state stmt (self#thisEffect ());
(*    rend_time "##### LOOP_after_upd" rtime*)
   (* setInstrEffect s
        (self#thisEffect ())*)


(*log "\nSucc (%d) = %d\n" cnt (List.length stmt.succs);
List.iter (fun s ->log "\nSucc (%d):\n" cnt; 
                   let _ = self#doStmt s in ()) stmt.succs*)
  end
 done;
 logtxt  "\n}\n"; 
 (* check if all leafs yield the same effect*)
 let effectList (sid:int) (el:effect list) : effect list =
  match absFind IntMap.find sid sid2effect with
    | Some (_,Some f,_) -> 
        el @ [f]
    | _ ->  
       raise (Impossible 
       ("effect_of_leaf (sid=" ^ (string_of_int sid)^")")) 
 in 
 let el = IntSet.fold effectList leafs [] in
 if not (arePathsEqual effect2str  el) then
 begin
   err "Some return paths are not equal.";
 end;
 let outf = branch_of_paths el in
 outf

(*******************************************)
method doFundec (f:Cil.fundec) (args:  varinfo list): effect =
 (*count <- 0;*)
 edge_count <- 0;
 cont_sid  <-  IntMap.empty;
 sid2effect <- IntMap.empty;
 visited <- IntSet.empty;
 leafs <- IntSet.empty;
 (let fvi = str2varinfo f.svar.vname in
  let loc = fvi.vdecl in     (*fun_fundec.svar.vdecl in *)
  context_loc <- loc
 );
 context_fundec <- f ; 
 rec_effect <- ref None ;(*recursive effect*)
 resetLockVid ();

 if context_fundec.svar.vname = "main" then
  main_vid <- context_fundec.svar.vid;

 List.iter (*add formals to function summary*) 
 (fun vi -> self#addSum_aLval 
            (Lvals.abs_of_lval ((Var vi,NoOffset)))) args;

  match f.sbody.bstmts with
  | [] -> (* nothing to do -- empty function *) []
  | first_statement :: _ ->
    (* Calculate function effect *) 
    let effect0 = self#doStmt first_statement in

     (*log "\nOUTPUT_EFFECT of %s BEFORE: %s\n\n"
     (fundec2str context_fundec)
     (effect2str effect0);*)

    let effect = sgarbage effect0 in 
    effect_map <- IntMap.add f.svar.vid
    (f,effect,args) effect_map;
 
  (*  log "\nOUTPUT_EFFECT of %s AFTER: %s\n\n"
     (fundec2str context_fundec)
     (effect2str effect);*)
 
    calc_counts f.svar.vid effect;
    effect
end  
(**********************************************************
 *
 *        Context-sensitive effects
 *
 *
 **********************************************************)
 (*let (call_stack : (Cil.fundec * int) Stack.t) = Stack.create ()*)

    
 (* an effect whose variables are uniquely identified 
  * by a context
  * *)
 type ueffect = ctx * effect
 type stack = ctx * (ueffect list) 
 type lockset = uvid list
 type ctx_lockset = ctx * lockset
 type ctx_map = (ctx,lockset) Hashtbl.t
 type loc_ctx_map = (Cil.location,ctx_map) Hashtbl.t


 let (cont_effect: loc_ctx_map ) = Hashtbl.create 117

(* remove duplicates   *)
let lockset_rem_dups (l:lockset) : lockset =
 let eq (ctx1,vi1) (ctx2,vi2) = 
   let b =
   if vi1.vid <> vi2.vid then false
   else ctx_eq ctx1 ctx2
   in
(*  log "comparing %s@%s (vid=%d) AND %s@%s (vid=%d)\n"
          (varinfo2str vi1) 
          (ctx2str ctx1)
          (vi1.vid) 
          (varinfo2str vi2)
          (ctx2str ctx2)
          (vi2.vid); *)
      b 
 in
 let f hd tl =  List.exists (eq hd) tl  in
 let rec hdtl l = 
   match l with
   | hd::tl -> 
     let tl' = hdtl  tl in
     if f hd tl then tl'
     else hd::tl'
   | []  -> []

 in hdtl l

 (* add a lockset that corresponds to a specific 
  * lock operation to its lockset
  * *)
 let add_cont (loc:Cil.location) (ctx,lset)  : unit = 
 let ctx_tbl =
   match absFind Hashtbl.find cont_effect loc with
    | Some ctx_tbl ->
      ctx_tbl
    | None    -> 
      let ctx_tbl = Hashtbl.create 117 in
      Hashtbl.add cont_effect loc ctx_tbl;
      ctx_tbl
 in
 match absFind Hashtbl.find ctx_tbl ctx with
  | Some lset' ->
    let lset'' = lockset_rem_dups (lset @ lset') in
    Hashtbl.replace ctx_tbl ctx lset''
  | None -> 
    Hashtbl.add ctx_tbl ctx lset


 (* iterate over the elements of the context-sensitive
  * lockset table
  * *)
 let iter_cont (f:Cil.location->ctx_lockset->unit):unit = 
  let iter0 loc ctx lset = f loc (ctx,lset) in
  let iter1 loc ctx_tbl=Hashtbl.iter (iter0 loc) ctx_tbl in
  Hashtbl.iter iter1 cont_effect


 (* string conversion functions *)
 let ctx2str (ctx:ctx) =
    abs_list2string "->" (fun (fd,loc) ->
                  (fundec2str fd) ^ "@" ^
                   (loc2str loc )) ctx
 let lockset2str (l:lockset) : string =
   let f (ctx,vi) = "(" ^ (ctx2str ctx) ^ "," ^
                       (varinfo2str  vi) ^ ")" in
   abs_list2string "\n" f l



 (* compute lockset for this context *)
 let rec lockset (s:sexp) (n:int ref) (f:effect):lockset = 
  let doLockset (l:lockset) a : lockset=
   if !n < 1 then l 
   else begin
     match a with
      | LockOp((loc,_,_) ,b,s') -> 
        let n' = if b then 1 else (-1) in
        
        if sexp_leq s s' then
         begin
          n := !n + n';
          l
         end
        else if b then
          l@[uvid_of_sexp s']
        else
          l
      | Join el ->
        let n' = ref 0 in
        let (ls : lockset) =  List.fold_left
         (fun ls f' -> 
              n':=!n ; 
              let ls' = (lockset s n' f') in
              (*log "adding path lockset :%s\n"
              (lockset2str ls');*)
              ls @ ls'
         ) []  el 
        in
        n := !n';
        let l' = l @ ls in
        (*log "\nOUT_LOCKSET = %s\n\n\n"  (lockset2str l');*)
        l'
        
      | NCall (_,res_f) ->
        let f' = must_resolved "lockset/NCall" res_f in
        l @ (lockset s n f')

      | RCall _ -> 
            raise (Not_implemented "lockset/RCall")
   end        
  in     
  List.fold_left doLockset [] f 

 (* @acquire_effect_ctx: 
  *    registers the locksets for each context and
  *    lock operation.
  * *)
 let rec acquire_effect_ctx
(*  (map: (Cil.fundec*effect*Cil.varinfo list) HashMap.t)*)
  ((context,effect) : stack) (f:effect) : unit =
  let doTail  =  acquire_effect_ctx (context,effect) in 
  match f with
  | LockOp((loc,_,_) as ci,true,s)::tl -> (*acquire operation*)
    (*let effect' = ([],f)::effect in*)
    (*let f = (context, effect') in*)
    let (f':effect) =tl @ ( List.fold_left
          (fun (f':effect) ((_,f):ueffect)->f'@f ) 
             [] effect) in
    let n' = ref 1 in 
    let lset0 =(* (log "LOCKSET at %s OF EFFECT %s WITH TAIL= %s\n\n"
                     (loc2str loc) (effect2str f')
                     (effect2str tl)
                );*)
                lockset s n' f' in
    let lset = lockset_rem_dups lset0 in
    add_cont loc (context,lset);
    

    let (ctx:string) = ctx2str context in
   
    log "Acquire operation at %s\nContext:\n%s\nEffect (f'=%d):\n%s\n-----------------------\n\n\nLockset (n'=%d,lset=%d):\n%s\n-----------------------\n"
    (loc2str loc) (ctx) (List.length f')
    (effect2str_detail true false (LockOp(ci,true,s)::f'))
    (!n') (List.length lset)
    (lockset2str lset);

    doTail tl
     
  | LockOp _ ::tl ->   (* release operation *)

    doTail tl

  | (Join el)::tl -> 
    List.iter 
    (acquire_effect_ctx (context, (context,tl) :: effect)) el;

    doTail tl

  | NCall((loc,_,vi),f_res)::tl ->
    let vid = vi.vid in
    begin match vid2fundec_opt vi.vid with
    | Some fd ->
      let f = must_resolved "call_effects" f_res in
      log "\n======================\nFUNCTION:%s\n======================\n\nFUN_EFFECT:\n%s\n@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n\n" 
         (vid2str vid) (effect2str f);
      let ctx' = context@[(fd,loc)] in
       acquire_effect_ctx (ctx',(context,tl) :: effect) f
    | None  ->
      ()
    end; 
    doTail tl

  | RCall _::_ -> 
        raise (Not_implemented "call_effects")

  | [] -> ()
  
(**********************************************************
 *
 *    More Effect Optimizations 
 *
 *
 **********************************************************)

(* check whether there exist any top-level lock-ops 
 * or any lock ops that WILL access the stack frame
 * *)
let rec should_gen_code (f:effect) : bool = 
 List.exists (function 
  | LockOp _ -> true
  | Join el -> List.exists should_gen_code el
  | NCall ((_,_,vi),_) -> will_backtrack vi.vid
  | RCall _ -> false
 ) f 
 

(* gather lockops*)
let rec sgather (f:effect) : (effect * effect) =
  match f with 
   | [] -> ([],[])
   | LockOp a :: tl ->
     let (b,c) = sgather tl in
     (LockOp a :: b, c)
   |  _  -> ([],f)

(* *)
let rec seffect (f:effect) : effect =
 log "\nSEFFECT : %s\n" (effect2str f);
 match f with 
 | [] ->   [] 
 | RCall _ :: _ -> 
   raise (Not_implemented "seffect/RCall") 
 | NCall _:: _ -> (* must be simplified*) 
   raise (Impossible "seffect/NCall") 

 | _ -> 
   let (lkops,join) = sgather f in
   log "\nS@@@ lockops: %s\n\njoin:\n%s\n\n"
   (effect2str lkops) (effect2str join);
   begin match join with
   | [] -> lkops
   | Join el :: tl ->
     let (lkops',join') = sgather tl in
     log "\nS@@@ lockops': %s\n\njoin':\n%s\n\n"
     (effect2str lkops') (effect2str join');
     let tr f = seffect (lkops @ f @ lkops') in
     let el' = List.map tr el in
     (Join el')::(seffect join')
   | _ -> 
     let str = effect2str join in
     raise (Impossible ("seffect/default case" ^ str)) 
   end


(*
*)
(**********************************************************
 *
 *        Code Generation
 *
 *
 **********************************************************)
(* Field creation *)
let mkfield s f ci = 
   (s,f ci,None,[],Cil.locUnknown ) 


let mkrecfield s  =  (* recursive field*)
   mkfield s (fun ci-> TPtr(TComp (ci,[]),[])) 

let mksfield s t = (* simple field*)
   mkfield s (fun ci -> t) 

let mkstruct s flds =  (* create a struct*)
  mkCompInfo true s
    (fun ci ->
       List.map (fun x -> x ci) flds
    ) []

(**************************************************)
(**********************************************************
 *
 *        Intra-procedural Effects
 *
 *
 **********************************************************)

(***************************************
 *  get type of struct  node_t 
 ***************************************)
let max_effect = 255
let max_offset = 255
let max_parent = 255

(**)
let rec get_ci t  = 
 match t  with 
 | TComp(ci,_) ->  ci
 | TNamed(ti,_) -> get_ci ti.ttype
 | TPtr(t',_) ->  get_ci t'
 | _ -> 
   rs_impossible "ci_of_node: %s" (typ2str t)


(**)
let fieldinfos = Hashtbl.create 10

(**)
let init_fieldinfos t = 
 let add f s =  Hashtbl.add fieldinfos s (f s) in
 let node_fld = Cil.getCompField (get_ci t) in

 (*A. meta info *)
 let meta_info = (node_fld "meta_info")  in
 let meta_info_fld =  Cil.getCompField (get_ci meta_info.ftype) in

 let kind = (node_fld "kind")  in
 let kind_fld =  Cil.getCompField (get_ci kind.ftype) in

 let add_meta_info = add (fun s->  meta_info_fld s) in
 let add_kind = add (fun s->  kind_fld s) in

 (* Fill in the hashtable *)
 add (fun _ -> meta_info)   "meta_info" ;
 add (fun _ -> kind )       "kind"      ;

 add_meta_info "tag" ;
 add_meta_info "type" ; 
 add_meta_info "mem" ;
 add_meta_info "has_parent" ;
 add_meta_info "dummy" ;
 add_meta_info "parent" ;
 add_meta_info "parent_offset" ;
 add_meta_info "size" ;

 add_kind "stack_offset";
 add_kind "handle";
 add_kind "com";

 try
   add (fun s -> node_fld s) "node_name" 
 with _ -> () 


(*get field info *)
let get_fi s =
  try 
   Hashtbl.find fieldinfos s
  with Not_found -> 
   rs_impossible "Field %s was not found !\n" s

(**)
let stack_struct_name = "stack" 

let rstack_vi_opt = ref None
let rstack_vi () : varinfo =
 match !rstack_vi_opt with
 | None -> 
   begin match varinfo_of_name stack_struct_name with
    | Some ret -> 
      rstack_vi_opt := Some ret;
      ret
    | None -> 
      rs_impossible "rstack_vi" 
   end
 | Some r ->  r

(**)
let stack_typ () = 
   match (rstack_vi ()).vtype with
    TPtr(t,_) -> t
   | _ -> 
      rs_impossible "stack_typ : %s" 
                    (typ2str ((rstack_vi ()).vtype))

(**)
let typ_node_t_opt = ref None 
(* this function should have been a lot easier to write
 * if it wasn't for the stupid RELAY stuff*)
let typ_node_t () : Cil.typ =
 let failwith t n = 
   let p = any2str d_plaintype t in
     raise (Impossible 
            ("typ_node_t/" ^ n  ^ ": " ^ p))
 in
 match !typ_node_t_opt with 
 | None -> 
     let vi = rstack_vi () in
     begin match vi.vtype with 
      | TPtr(TComp(ci,_),_) -> 
        let current = Cil.getCompField ci "current" in
        let (node_t:typ) = current.ftype in

        Hashtbl.add fieldinfos "current" current ;
        begin match node_t with
        | TPtr(TNamed(ti,_),[]) -> 
          typ_node_t_opt := (Some ti.ttype);
          init_fieldinfos ti.ttype;
          ti.ttype (* struct node *)
        | _ ->
          failwith node_t "1"
        end
      | _ ->
         failwith vi.vtype "2"
     end 
 | Some x -> (x:Cil.typ)
(*******************************)
let mkptr_typ t = 
   TPtr(t,[]) 


(**)
let vid2vi = Hashtbl.create 100

let mk_fun_effect vid : varinfo =
 let node_t = typ_node_t () in
 let typ_node_t_star = mkptr_typ node_t in
 (*mkarray_typ node_t (-1)  in*)
 let name = ("__effect_" ^ (string_of_int vid)) in
 let vi =  makeGlobalVar name typ_node_t_star in
  vi.vstorage <-  Extern;
  Hashtbl.add vid2vi vid vi;
  vi

let fun_effect vid : varinfo = 
(  try 
     Hashtbl.find vid2vi vid 
  with _ -> 
     mk_fun_effect vid
)

(*******************************)

let mkarray_typ t sz = 
   TArray (t, 
         (if sz < 0 then None 
         else Some (Cil.integer sz)),[])

let ptr_of_typ t =  TPtr(t,[]) 

let mktemp_var fd typ = makeTempVar fd typ  

let mkframe_var fd = 
 mktemp_var fd (stack_typ ())

let mkpthreadarray_var fd sz = 
 let t = ptr_of_typ (get_fi "handle").ftype in
 mktemp_var fd (mkarray_typ t sz)

(* add the new variables to the function definition and
 * return a set of instructions that should be inserted
 * at the begging of the function specified by fd
 * *)
let mk_rtstack fd (vlist: lval list):
                     (Cil.varinfo * Cil.instr list) =
                           
 let arr = mkpthreadarray_var fd (List.length vlist)  in
 let frame = mkframe_var fd in
 let set i vi' : Cil.instr = 
  let rhs = Cil.mkAddrOf vi' in
  let lhs = (Var arr, Index (Cil.integer i,NoOffset)) in
  (Set (lhs,rhs, Cil.locUnknown))
 in
 let cnt = ref (-1) in
 let init_arr =
    List.map (fun vi -> 
               cnt:=!cnt+1;
               (*Hashtbl.add var_map vi !cnt;*)
               set !cnt vi) vlist in
 (* end initialize array
  *
  * now init stack frame
  * *)
 log "STACK NODE TYPE: %s\n" (typ2str (stack_typ ()));
 let  ci_stack_node = get_ci (stack_typ ()) in
 let ci_stack_fld = Cil.getCompField ci_stack_node in
 let next_field =  ci_stack_fld "next" in
 let locals_field =  ci_stack_fld "locals" in
 let current_field =  ci_stack_fld "current" in
(* let base_field =  ci_stack_fld "base" in*)
 (* frame->next = rstack *)
 let rstack_offset = NoOffset in
 let next_offset = Field (next_field,NoOffset) in   
 let locals_offset = Field (locals_field,NoOffset) in   
 let current_offset = Field (current_field,NoOffset) in   

(* let base_offset = Field (base_field,NoOffset) in   *)

 let lhs0 = (Var frame, next_offset) in
 let rhs0 = Lval (Var (rstack_vi ()),rstack_offset) in
 let i0 =  (Set (lhs0,rhs0, Cil.locUnknown)) in

(* let lhs2 = (Var frame, base_offset) in
 let rhs2 = Lval (Var ef_vi,NoOffset) in
 let i2 =  Set (lhs2,rhs2, Cil.locUnknown) in*)
 (* rstack = &frame *)
 let lhs1 = (Var (rstack_vi ()),rstack_offset) in
 let rhs1 = Cil.mkAddrOf (Var frame,NoOffset) in
 let i1 =  (Set (lhs1,rhs1, Cil.locUnknown)) in

 let lhs2 = (Var frame,locals_offset) in
 let rhs2 = Lval (Var arr,NoOffset) in
 let i2 =  (Set (lhs2,rhs2, Cil.locUnknown)) in

 (* stack.current = __effect_x *)
 let fun_vid   = fd.svar.vid in 
 let effect_vi = fun_effect fun_vid in
 let lhs3 = (Var frame, current_offset) in
 let rhs3 = Lval (Var effect_vi,NoOffset) in
 let i3 =  (Set (lhs3,rhs3, Cil.locUnknown)) in
 (* compose frame instructions*)
 let init_frame = [i0;i2;i1;i3] in 
 (* compose all instructions *)
 (frame, init_frame @ init_arr)



(***)
let pop_stack () :  Cil.instr  =
 let frame_ci = get_ci (stack_typ ()) in
 let fieldinfo = Cil.getCompField frame_ci "next" in
 let offset = Field (fieldinfo,NoOffset) in   
 let rs_offset = NoOffset in
 let rstack_exp = Lval (Var (rstack_vi ()),rs_offset) in
 let rhs0 = Lval (Mem rstack_exp, offset) in
 let lhs0 = (Var (rstack_vi ()), NoOffset) in
 let i0 =  (Set (lhs0,rhs0, Cil.locUnknown)) in
 i0
(*******************************)
(**)
let arrayi vi ioffset : lval =
 let idx = Index (Cil.integer ioffset,NoOffset) in
 let lhs = ((Var vi,idx)) in
 lhs

 (*use : addOffset end current 
  *      addOffsetLval o lval 
  * *)

(**)
let array_set l o rhs : Cil.instr =
 Set (addOffsetLval o l,rhs,Cil.locUnknown)

(**)
(*if debugging is off then this field does not exist*) 
let set_name l name_v : Cil.instr list = 
 try
  let name = get_fi "node_name" in
  (Set (addOffsetLval (Field (name,NoOffset)) l,
      (Const (CStr name_v )),Cil.locUnknown)
  )::[]
 with _ -> [] 

(*tokenize a string*)
let rec split_char sep str =
 try
  let i = String.index str sep in
    String.sub str 0 i ::
    split_char sep (String.sub str (i+1) 
    (String.length str - i - 1))
 with Not_found -> [str]

(* convert a string of the form "field1.field2.field3"
 * to a Cil offset
 * *)
let s_o str = 
 let strs = split_char '.' str in
 List.fold_left (fun o s -> 
    addOffset (Field (get_fi s,NoOffset)) o
 ) NoOffset strs

(**)
let mkarray_typ t sz = 
   TArray (t, 
         (if sz < 0 then None 
         else Some (Cil.integer sz)),[])

(**)
let glob_init_instr = (Hashtbl.create 100:
                       (int,instr list) Hashtbl.t)
   (*ref [] *)
(**)
let get_glob_instr vid = 
 match  absFind  Hashtbl.find glob_init_instr vid with
 | Some x -> x 
 | None -> [] 

let add_glob_init vid elt =
  Hashtbl.replace glob_init_instr
           vid ((get_glob_instr vid) @ elt)

let finalizeGlobInit () =
 let fd = getGlobInit (opt2some !cilfile) in
 let i = Hashtbl.fold 
 (fun _ instr ilist -> ilist @ instr )
 glob_init_instr []  in
 fd.sbody.bstmts <- [Cil.mkStmt (Instr i)]

(* the following defs must be kept in sync with
 * the header file *)
type mem = 
   | STACK of int
   | GLOBAL of sexp
   | HEAP   of sexp

(**)
let mem2i m = 
 match m with
 | STACK _ -> 0
 | GLOBAL _ -> 1
 | HEAP  _ -> -2

(**)
let mem2strexp mem = 
 begin match mem with
 | STACK o ->
   ("stack_offset", (integer o))
 | HEAP _ ->
   rs_not_implemented "mem2strexp" 
 | GLOBAL s ->
   ("handle", sexp2exp s)
 end

(**)
type lock = 
   | LOCK
   | UNLOCK

(**)
let lock2i l = 
 match l with 
 | LOCK -> 1
 | UNLOCK -> (-1) 

(**)
let bool2lock b =
  if b then LOCK else UNLOCK 

(**)
(**)

type tag =
   | SIMPLE of string * mem * lock
   | SEQ  of varinfo * int
   | PATH of varinfo * int
   | CALL of (varinfo * int) option * string 

(**)
let tag2strexp tag  =
let addo vi o =
 let lval_e = Lval (Var vi,NoOffset) in
  ("com",Cil.increm lval_e o)
in match tag with
 | SIMPLE (_,m,_)-> mem2strexp m 
 | SEQ (vi,i) -> addo vi i 
 | PATH (vi,i) -> addo vi i 
 | CALL (o,s) -> 
   begin match o with
   | Some (vi,i) -> addo vi i
   | None -> 
      let null = CastE (TPtr (typ_node_t (),[]), zero) in
      ("com",null)
   end 

(**)
let tag2i k =
   match k with
   | SEQ _ -> -2
   | SIMPLE _ -> 0
   | PATH _ -> 1
   | CALL _ -> -1

(**)
let tag2str k =
   match k with
   | SEQ _ -> "SEQ"
   | SIMPLE (s,_,_) -> "SIMPLE (" ^ s ^ ")"
   | PATH _ -> "PATH"
   | CALL (_,s) ->  "CALL (" ^ s ^ ")"


(**)
let check_offsets lst = 
  List.iter (fun o -> if o > 255 then
             begin
                err "Offset > 255\n";
                rs_impossible "OFFSET IS BIG: %d\n" o
             end
 ) lst


(**)
let glob_init_static= (Hashtbl.create 100:
                       (int,(offset*init) list) Hashtbl.t)
   (*ref [] *)
(**)
let get_glob_init_static vid = 
 match  absFind  Hashtbl.find glob_init_static vid with
 | Some x -> x 
 | None -> [] 

let add_glob_init_static vid elt =
 (* log "add_glob_init_static vid = %d\n" vid;*)
  Hashtbl.replace glob_init_static
           vid ((get_glob_init_static vid) @ elt)

type node_init = int -> (*offset *)
                  (* abs parent, poffset array size *)
                 (tag *int *int *int) -> 
                 (offset*init) list

(*
 * code for static initialization of nodes*)
let node_init 
  offset  (tag,parent,poffset,size)  = 

 (*let ini fi i = (Field (fi,NoOffset), i) in*)
 let sim_inis s exp =(Field ((get_fi s),NoOffset),
                     SingleInit exp) in
 let comp_ini s b =
   let fi = get_fi s in
   (Field (fi,NoOffset),CompoundInit (fi.ftype,b)) 
 in

 let (kstr,kexp) = tag2strexp tag in

 let kind_ini = comp_ini "kind" [sim_inis kstr kexp] in
 let i ii = Cil.integer ii in
 let (has_parent,parent',poffset') =
  if parent < 0 then (0,0,0) 
  else (1,offset-parent,poffset) 
 in

 let (ilktyp,imem) = 
  match tag with 
  |  SIMPLE (_,mem,lktyp) ->
    (i (lock2i lktyp),i (mem2i mem))
  | _ ->
    (zero,zero)
 in


 let meta_info_ini = comp_ini "meta_info" 
                     [sim_inis "tag" (i (tag2i tag)) ;
                      sim_inis "type" ilktyp;
                      sim_inis "mem"  imem; 
                      sim_inis "has_parent" (i has_parent) ;
                      sim_inis "dummy"  zero ;
                      sim_inis "parent" (i parent');
                      sim_inis "parent_offset" (i poffset');
                      sim_inis "size" (i size)
                     ]  
 in 
 let node_t = typ_node_t () in
 let name = ((tag2str tag) ^ " @ " ^  (string_of_int offset)) in
 let node_ini = CompoundInit (node_t,
                [meta_info_ini;
                 kind_ini;
                 sim_inis "node_name" 
                           (Const (CStr name))
                ])
in 
 let idx = Index (Cil.integer offset,NoOffset) in
 [(idx,node_ini)]


(**)
let effect_init inis = 
 let node_t = typ_node_t () in
 let node_array_t = mkarray_typ node_t  (-1) in
 CompoundInit (node_array_t, inis)

(*
let node_array_t = mkarray_typ node_t  (-1) in
let vi = fun_effect vid in 

CompoundInit array_node_t  
 [ (idx0


SingleInit (Cil.integer (tag2i  kind_v))

( (Index 0,NoOffset), CompoundInit node_t [((Field meta_info,NoOffset), i1);

									       (Field kind, i2);
										(Field node_name, i3)
					 ; 
				      ((Index 1 ... ] 

in
basic_fun
*)



(**)
(* Common instructions for all kinds of 
 * effect nodes
 * *)
(* Example:
   n0[0].name = "k @ 0";
   n0[0].meta_info.tag = k;
   n0[0].meta_info.has_parent = 0;
   n0[0].meta_info.parent = 0;
   n0[0].meta_info.parent_offset = x; //optional
   n0[0].meta_info.size = 4;
*)
let make_effect
         (k:tag)
         (vid:int)
         (offset:int) 
         (parent:int)  (* parent node offset *)
         (poffset:int) (*this node's offset in parent array*)
         (nodes:int) :
         (varinfo * (string-> Cil.exp->unit)) =

 if offset < parent+1 then 
  rs_impossible "make_effect: offset=%d parent=%d" offset parent;
 (* check offsets *)
 check_offsets [offset;parent;poffset;nodes];
 (* get fields and definitions*)            
 let vi = fun_effect vid in 
 (* n0[0] *)
 let ai = arrayi vi offset in 
 let sai s e = add_glob_init vid 
                [(array_set ai (s_o s) e)] in
 (* .name = "tag @ " + offset *)
 let tail = set_name ai ((tag2str k) ^ " @ " ^
                          (string_of_int offset)) in
 let i k = integer k in
 let (has_parent,parent',poffset') =
    if parent < 0 then (0,0,0) else (1,offset-parent,poffset) in
 add_glob_init vid tail;
 sai "meta_info.tag" (i (tag2i k));
 sai "meta_info.has_parent" (i has_parent);
 sai "meta_info.parent" (i parent');
 sai "meta_info.parent_offset" (i poffset');
 sai "meta_info.size" (i nodes);
 (vi,sai)


(**)
(* Example:
  n0[0].name = "SEQ @0";
  n0[0].meta_info.tag = SEQ;
  n0[0].meta_info.has_parent = 0;
  n0[0].meta_info.parent_offset = x; //optional
  n0[0].meta_info.size = 4;
  n0[0].kind.com = n0+0+1;
 * *)
let make_seq 
         (vid:int)     (* vid of fun using this effect*)
         (offset:int)  (* array index*)   
         (parent :int) (* idx of parent *)
         (poffset:int) (* offset in parent array*)
         (nodes:int)   (* number of children *)
         (coffset:int) (* child offset*)
         : unit =
 (* get fields and definitions*)            
 let (vi,sai) = make_effect (SEQ (dummyFunDec.svar,coffset))
                vid offset parent poffset nodes in
 let lval_e = Lval (Var vi,NoOffset) in
 sai "kind.com"  (Cil.increm lval_e (coffset))

(**)
(* Example: 
  n0[3].name = "PATH @ 3";
  n0[3].meta_info.tag = PATH;
  n0[3].meta_info.has_parent = 1;
  n0[3].meta_info.parent = 3; //offset from parent
  n0[3].meta_info.parent_offset = 2;
  n0[3].meta_info.size = 2; // paths

  n0[3].kind.com = n0+3+2;
*)
let make_path 
         (vid:int)     (* vid of fun using this effect*)
         (offset:int)  (* array index*)   
         (parent :int) (* idx of parent *)
         (poffset:int) (* offset in parent array*)
         (nodes:int)   (* number of children *)
         (coffset:int) (* child offset*)
         : unit =
 (* get fields and definitions*)            
 let (vi,sai) = make_effect (PATH (dummyFunDec.svar,coffset))
                vid offset parent poffset nodes in
 let lval_e = Lval (Var vi,NoOffset) in
 sai "kind.com"  (Cil.increm lval_e (coffset))


(**)
(* Example: 
  n0[3].name = "CALL(function_name) @ 3";
  n0[3].meta_info.tag = CALL;
  n0[3].meta_info.has_parent = 1;
  n0[3].meta_info.parent = 3; //offset from parent
  n0[3].meta_info.parent_offset = 2;
  n0[3].meta_info.size = 1; 

  n0[3].kind.com = n1;
*)
let make_call
         (vid:int)     (* vid of fun using this effect*)
         (offset:int)  (* array index*)   
         (parent :int) (* idx of parent *)
         (poffset:int) (* offset in parent array*)
         (fun_vi:varinfo )
         (is_empty_effect : bool)
         : unit =
 let (_,sai) = make_effect (CALL (None,fun_vi.vname)) 
                vid offset parent poffset 0 in
 sai "kind.com"
 (if not is_empty_effect then        
  begin
   let call_vi = fun_vi.vid in 
   let effect_vi = fun_effect call_vi in 
   let lval_e = Lval (Var effect_vi,NoOffset) in
    lval_e  
  end 
  else 
     CastE (TPtr (typ_node_t (),[]), zero)
 )

(**)
(* Example:
 n0[1].name = "SIMPLE @1: &p0";
 n0[1].meta_info.tag = SIMPLE;
 n0[1].meta_info.has_parent = 1;
 n0[1].meta_info.parent = 1; //offset from parent
 n0[1].meta_info.parent_offset = 0;
 n0[1].meta_info.size = 3; //parent size

 n0[1].meta_info.type = LOCK; 
 n0[1].meta_info.mem = STACK; 
 n0[1].kind.stack_offset = OFF(3);

 *)

 let make_simple 
         (msg:string)
         (vid:int)     (* vid of fun using this effect*)
         (offset:int)  (* array index*)
         (parent :int) (* idx of parent *)
         (poffset:int) (* parent offset *)
         (nodes:int)   (* number of parent nodes*)
         (lockop:bool) (* lockop true=lock,false=unlock*)
         (mem:mem)
         : unit =
 (* get fields and definitions*)   
 let lockop_str = if lockop then "+" else "-" in
 let l = bool2lock lockop in
 let (_,sai) = make_effect (SIMPLE (msg ^ lockop_str,mem,l))
                vid offset parent poffset nodes in
 sai "meta_info.type" (integer (lock2i l));
 sai "meta_info.mem"  (integer (mem2i mem));
 match mem with
 | STACK o ->
   sai "kind.stack_offset" (integer o)
 | HEAP _ ->
   raise (Not_implemented "make_simple/heap") 
 | GLOBAL s ->
   sai "kind.handle" (sexp2exp s) 



(***************************************)
(*type fun_map =  (int,(int * (Cil.lval,int) Hashtbl.t)) Hashtbl.t
let fun_map = ref (None : fun_map option) 

(**)
let getFunMap () : fun_map =
 match !fun_map with 
  | None -> 
   let fname = Filename.concat "ciltrees" "funids.bin" in
   let ic = open_in fname in
   let fmap = (Marshal.from_channel ic : fun_map) in
   fun_map := Some fmap;
   close_in ic;
   fmap 
 | Some f -> f

(**)
let getFunMapFd fd_vid =
 let fm = getFunMap () in
 log "Listing hashtbl keys :\n" ;
 Hashtbl.iter
 (fun k _ -> 
    log "Key : %d\n" k
 )  fm;
 Hashtbl.find fm fd_vid 

(**)
let fd2effect_vid fd_vid = 
  fst (getFunMapFd fd_vid)

(**)
let fd2lvalmap fd_vid = 
  snd (getFunMapFd fd_vid)
*)
(***************************************************)
let rec effect_size (f:effect) : int =
 let len = List.length f in
 if len = 0 then 0
 else 
  let inner_sz =  List.fold_left (fun sz a ->
    sz+ 
    match a with
    | LockOp _ -> 1
    | Join el -> 
      List.fold_left (fun sz f0-> sz + (effect_size f0)) 1 el
    | NCall (_,res_f) -> 
      let f' = must_resolved "effect_size" res_f in
      1+(effect_size f') 
    | RCall _ -> 1
   ) 0 f  
  in
   inner_sz + (if len = 1 then 0 else len)


(**)
let mkEffectInstr (vi:varinfo)
                  (f:effect) : Cil.instr list = [] 

let no_effect_gen = ref IntSet.empty

let lockop_map = ref [] 

let lockop_offset loc = 
(* try*)
   snd (List.find (fun (loc',_) -> loc' = loc) !lockop_map)
(* with Not_found ->
   rs_impossible "Could not find location %s\n" (loc2str loc) *)
    
let next_pos = ref 0 

(**)
let doIntraEffect 
   (effectOf : int -> effect) fd  (f:effect) : unit =
 let fd_vid = fd.svar.vid in

 (*begin effect opt *)
 let f0 = sgarbage f in (*remove garbage effects*)
 (*log "BEFORE SHRINK : %s\n" (effect2str f0);*)
 let f1 = shrink f0 in

if should_gen_code f1 then
begin
 (*log "AFTER SHRINK : %s\n" (effect2str f1);*)
 let f2 = scall f1 in (*INVARIANT:  scall AFTER shrink ONLY*)

 let f3 = sgarbage f2 in
 let f' = shrink f3 in
 (*end effect opt *)
 

(*  log "Effect of function %s (vid=%d VID=%d):\n\nSGARBAGE=%s\n\nSCALL/SHRINK=%s\n\n"
      (vid2str fd_vid) fd_vid fd.svar.vid
      (effect2str_detail true false f0) 
      (effect2str_detail true false f') ;*)
  let fsize = effect_size f' in

  if fsize > 255 then 
  begin
     err "Effect too large: %d nodes" fsize;
     exit (-1)
  end;

  log "Effect of function %s (vid=%d VID=%d,SIZE=%d): %s\n\n"
      (vid2str fd_vid) fd_vid fd.svar.vid fsize
      ((*let doc = pretty_effect f' in 
        sprint 40 doc*)
         effect2str_detail true true f'
      );


   ;ignore(typ_node_t ()); (*INITIALZE type info 
                           DO NOT REMOVE
                          *)

  lockop_map := [] ; (* reset lockop map *)
  next_pos := 0 ; (*reset next_pos*)

 let add_lockop loc offset =
    lockop_map := (loc,offset)::(!lockop_map)  in
 let worklist = Queue.create () in
 let enq e = Queue.add e worklist in 
 let deq () = Queue.take worklist in
 let notEmpty () = not (Queue.is_empty worklist) in
 let fadd () = let n = !next_pos in 
                   next_pos := n+1; n  
 in
 let qoffset () = 
    let n = Queue.fold (fun o _ -> o+1) 0 worklist in
    n+1
    (*if n = 0 then 1
    else n *)
 in
 let is_globinit f =
  match f with
   | NCall((call_loc,_,called_vi),_) -> 
    (strstr called_vi.vname 
      "__globinit_cil" 0)    <> -1
   | _ -> false
 in
  
 (* Each tuple on the queue has type
  * effect * parent offset * my_offset_in_parent *
  * #parent_nodes
  * *)                   
 enq (f',-1,-1,0); (* enqueue effect *)
 while notEmpty () do
 let (f'',parent_offset,my_parent_offset,parent_nodes) = 
     deq () in

 match f'' with 
 | [] -> () 
 | hd::[] ->
   begin match hd with
   | LockOp((loc,_,_),is_acquire,sexp) ->
     (*FIXME: assuming that sexp is global*)         
     let sexp_vid = vid_of_sexp sexp in
     let vi = Cilinfos.getVarinfo sexp_vid in 
     let offset = fadd () in
     let mem = if vi.vglob then 
                   GLOBAL sexp
                  else
                   STACK (-1) (*FIXME: no support for stack
                                       or heap variables
                               *)
     in
     let descr = ("|"^(loc2str loc)^"|"^(sexp2str sexp)) in
     (*log "SIMPLE %d = %s\n" offset (atomic_effect2str hd);
      *)
     add_lockop loc offset; (* we need this for the final
                               pass 
                             *)

     make_simple descr fd_vid offset
                 parent_offset my_parent_offset
                 parent_nodes is_acquire  mem;

    let (lk,lks) = if is_acquire then (LOCK,"+") 
                   else (UNLOCK,"-") in
    (*static initializer *)
    add_glob_init_static fd_vid 
    (node_init offset 
       ((SIMPLE (descr ^ lks,mem,lk)),
       parent_offset,my_parent_offset,parent_nodes
      )
    )

   | Join el ->
    begin match el with
     | [] -> () 
     | _  -> 
      let eff_vi = fun_effect fd_vid in
      let len = List.length el in
      let offset = fadd () in 
      let q = qoffset () in
      let qo = offset + q in

      (*log "\nPATH : %d qoffset=%d offset+qoffset=%d\n"
      offset q qo;*)

      make_path fd_vid offset 
                parent_offset my_parent_offset
                len qo ;

      add_glob_init_static fd_vid 
      (node_init offset (PATH (eff_vi,qo),
                        parent_offset,my_parent_offset,len));
 

      ignore(List.fold_left
      (fun i path -> 
         enq (path,offset,i,len);
         i+1
      ) 0 el)
    end

   | NCall((call_loc,_,called_vi),_) -> 
     let r = (strstr called_vi.vname 
                "__globinit_cil" 0) in
     if r  = -1 then
     begin
      (*let f' =  effectOf called_vi.vid in*)
         (*must_resolved 
                "doIntraEffect/NCall" res_f in*)
      let offset = fadd () in 
      (*log "CALL %s : parent_offset= %d\n" 
      (called_vi.vname) my_parent_offset;*)

      make_call fd_vid offset parent_offset
                 my_parent_offset 
                 called_vi true;

      add_glob_init_static fd_vid 
      (node_init offset (CALL (None,called_vi.vname),
                         parent_offset,my_parent_offset,0));


     (* we need this for the final pass *)
     add_lockop call_loc offset
    
     end
(*   | NCall _ *)
   | RCall _ -> 
      raise (Impossible
              ( "doIntraEffect/1 ==>  " ^
                 (effect2str_detail true false  (hd::[]))
              )
            )
   end
 | _ -> (*at least two elements*)
   let eff_vi = fun_effect fd_vid in
   let len = List.length f'' in
   let offset = fadd () in 
   let q = qoffset () in
   let qo = offset + q in
   (*log "\nSEQ : %d qoffset=%d offset+qoffset=%d\n"
   offset q qo;*)
   make_seq fd_vid offset parent_offset
            my_parent_offset len qo;

   add_glob_init_static fd_vid 
      (node_init offset (SEQ (eff_vi,qo),
                         parent_offset,my_parent_offset,len));

   ignore(List.fold_left
      (fun i elt ->
         if is_globinit elt then i
         else begin 
            enq ([elt],offset,i,len);
            i+1
         end
      ) 0 f'')

 done (*end while *)
end else begin
   no_effect_gen := IntSet.add fd_vid !no_effect_gen;
   log "Effect of function %s (vid=%d VID=%d): %s NO CODE GEN\n"
      (vid2str fd_vid) fd_vid fd.svar.vid
      (effect2str_detail true true f1);

end

 (*
  * of  
    (* location of call, location of statement, 
     * function that is called
     * argument to lock op 
     * bool flag acquire/release 
     * (aExp = aliased formal param or allocation point or global)
     * *)
   (lockopinfo * bool * sexp)

| Join  of effect list 

  *)

(**********************************************************
 *
 *        Insert Effect Offsets
 *
 *
 **********************************************************)

class insertVisitor = object (self) inherit Cil.nopCilVisitor as super 

val mutable call_loc      =  Cil.locUnknown 
val mutable vid           = -1
val mutable current_vi    = Cil.dummyFunDec.svar
val mutable called_vi     = Cil.dummyFunDec.svar

method private doCall e =
begin match e with
 | Lval( (Var(f),_) (*as lv*) ) -> 
  called_vi <- f;
  begin match f.vname  with
   | "mypthread_mutex_unlock" ->
         self#doRelease () 
   | "mypthread_mutex_lock" ->  
         self#doAcquire () 
   | _ ->  self#doNormal () 
  end
  (* Indirect call *)
 | Lval( (Mem(Lval(Var(f),_)),_) (*as lv*)) -> 
   called_vi <- f;
   self#doIndirect ()
 | _ ->   
   self#doIndirect ()
end 

method doInsert () =
 try  
   let offset = lockop_offset call_loc in
   let effect_vi = fun_effect vid in
   let lval_e = Lval (Var effect_vi,NoOffset) in
   let rhs =   (Cil.increm lval_e offset) in
(*   let rstack_vi = get_rstack_vi () in*)

   let lhs = ( Var current_vi
               (*Mem (Lval(Var rstack_vi,NoOffset))*),
              Field (get_fi "current",NoOffset)) in
   let set = Set (lhs,rhs,Cil.locUnknown) in
   (* log "INSERT : %s\n" called_vi.vname; *)
   self#queueInstr [set]
 with 
  | Not_found -> 
        (*log "\nNOT found for function :%s\n"
            (called_vi.vname); *)
        ()


method doNormal  ()  =
 let r = (strstr called_vi.vname 
                "__globinit_cil" 0) in
  if r  = -1 then   self#doInsert ()

method doAcquire  () = self#doInsert ()

method doRelease  () = () 
method doIndirect () = self#doInsert ()
(*
err "Don't have function pointer analysis yet.\
     Sorry, but indirect calls are disallowed."*)

(*val mutable pre = ([] : global list) 
val mutable post = ([] : global list) *)

val file2prepost = (Hashtbl.create 10 :
                    (string,(global list * global list))
                    Hashtbl.t)

method private addGlob () =
let sz = !next_pos in
let lst = get_glob_init_static vid in  (*to disable static init lst=[]*)
let lst_len = (List.length lst) in
if lst_len <> sz  then
    rs_impossible 
    "addGlob lst_len = %d sz=%d vid=%d" lst_len sz vid
   (* sanity check *)
else if sz > 0 then 
begin 
 let node_t = typ_node_t () in
 let ntyp = mkarray_typ  node_t sz in
 let vi = fun_effect vid in
 let glob_vi = makeGlobalVar vi.vname ntyp in
 let g2 = GVar (glob_vi,{init = Some (effect_init lst)},Cil.locUnknown) in
 let g1 = GVarDecl(vi,Cil.locUnknown) in
 vi.vtype <- ntyp;
 vi.vstorage <- Extern;

 let name =  (opt2some (!cilfile)).fileName in
 (*log "Adding globs for file : %s\n" name; *)
 begin  match absFind Hashtbl.find file2prepost name with
 | Some (pre,post) ->
    Hashtbl.replace file2prepost 
                     name (pre @ [g1], post @ [g2])
 | None -> 
    Hashtbl.add file2prepost name ([g1], [g2])
 end
end

method effectDecls name = 
begin  match absFind Hashtbl.find file2prepost name with
 | Some a -> a
 | None -> ([],[])
  (* rs_impossible 
     "Could not find globals for file: :%s\n" name*)
end 

(*List.split l*)
method vinst i = 
 begin match i with
 | Call(_,e,_,loc) -> 
  call_loc <- loc;
  self#doCall e
 | _ -> 
     ()
 end 
 ;
 SkipChildren

method vstmt s = 
 (match s.skind with 
  | Return _ ->
       self#queueInstr [pop_stack ()]
  | _ ->  () 
  );
  DoChildren

method vfunc fd =
 (* No effect generation for this function 
    if its id \in no_effect_gen set
  *) 
 vid <- fd.svar.vid;
 let no_insert_cond = (IntSet.mem vid !no_effect_gen)
   (*FIXME: calls must be "complete" *) 
 in
 if no_insert_cond  then
 begin
    log "NO CODE GEN for : %s\n"  fd.svar.vname;
    SkipChildren
 end
 else begin
   log "CODE GEN for : %s\n"  fd.svar.vname;
   let (frame_vi,vl) =  mk_rtstack fd [] in
   let si  = (Cil.mkStmt (Instr vl)) in
   fd.sbody.bstmts <- si::fd.sbody.bstmts ;
   current_vi <- frame_vi;
   self#addGlob (); (* add globals if necessary*)
   super#vfunc fd (* insert pop statements and offsets*)
 end
 (*
 ignore(List.exists 
 (fun vi ->
   let chk n n' = 
     if n = n' then
     begin
       current_vi <- vi;
       true
     end else false
   in
   match vi.vtype with
   | TNamed(ti,_) -> chk ti.tname "stack_node_t" 
   | TComp(ci,_) -> chk ci.cname "stack_node" 
   | _ -> false 
 ) fd.slocals); (* exists will terminate when the elt is found*)*)
end

let ivis = new insertVisitor 

let doInsertOffsets fn = 
 (* log "VFUNC : %s : %d\n" (fn.svar.vname) 
      (List.length  (fn.sbody.bstmts) );*)
  let cvis = ( ivis :    insertVisitor :>Cil.cilVisitor)  in
  let fn' =  visitCilFunction cvis fn  in
  fn'.sbody <- fn.sbody

(**********************************************************
 *
 *      Remove Scopes 
 *
 *
 **********************************************************)

class scopeVisitor = object (self)
   inherit Cil.nopCilVisitor as super 

method vvdec vi =  
 begin try
    Scope.removeScope vi
 with BadScope -> () end;
 SkipChildren

end

let doRemoveScopes file = 
 (* log "VFUNC : %s : %d\n" (fn.svar.vname) 
      (List.length  (fn.sbody.bstmts) );*)
  let svis = (new scopeVisitor :
                      scopeVisitor :>Cil.cilVisitor)  in
  visitCilFile svis file  

(**********************************************************
 *
 *        Finalization
 *
 *
 **********************************************************)
let mkpath  = Filename.concat 
let split_path f = 
   (Filename.dirname f, Filename.basename f)

(**)
let doFinalize dynamic_init f : unit =
 let is_the_typedef g = 
    begin match g with
    | GType (ti,_) ->
       ti.tname = "node_t" 
    | _ ->
      false 
    end 
 in
 let (pre,post) = ivis#effectDecls f.fileName in

 let decide hd tl = 
   let gname  = function
   | GVarDecl (vi, _) -> vi.vname
   | GVar (vi, _, _) ->  vi.vname
   | _ -> "" 
   in

   let is_gcc_builtin  = function
   | GVarDecl (vi, _) -> 
     let b1 = (strstr  vi.vname "__sync_" 0) = 0 in
     let b2 = vi.vstorage = Extern in
     let b3 = List.exists
             (function 
                | Attr(an,[]) -> 
                  an = "missingproto"
                | _ -> false
             ) (typeAttrs vi.vtype) in
     (*log "CONSIDERING : %s b1=%b b2=%b b3=%b\n" 
     vi.vname b1 b2 b3;*)
     b1 && b2 && b3   
   |  _ -> false
   in

  if (is_malloc_global (gname hd)) || 
     (is_gcc_builtin hd)  then
  begin
     tl
  end
  else hd::tl
 in

 let rec insert  lst =
  match lst with
  | [] -> []
  | _::[] -> lst
  | hd :: tl ->
    if is_the_typedef hd then
  (*  let vi  = opt2some (varinfo_of_name "main") in
    let ini1 = node_init 0 
                 (SEQ, LOCK, STACK 0,-1,0,0) in
    let ini2 = node_init 1
                 (SEQ, LOCK, STACK 2,-1,0,0) in
    let enit = effect_init (ini2 @ ini1) in*)

     (hd
        (* ::(GVar ( (fun_effect vi.vid), 
                {init = Some enit
                },
                 Cil.locUnknown
              ))*)
        ::pre) @ (insert tl)
    else
     decide hd (insert tl)
  in

    log "Finalizing instrumented file %s\n" f.fileName;
       
   (*let f = opt2some (!cilfile) in*)
    (*insert global effect definitions *) 
    f.globals <- (insert f.globals) @ post;

    (* place effects in glob init*)
     if dynamic_init then
       finalizeGlobInit ();

     (*remove scope annotations from file*)
     doRemoveScopes f;

    (* save to output file*)
    (let dir = mkpath "ciltrees" "gen" in
     let (old_path,fname) = split_path f.fileName in
     let fl =  mkpath dir  fname  in
     let oc = open_out  fl  in
     dumpFile defaultCilPrinter oc fl  f;
     close_out oc
    )


(**********************************************************
 *
 *        INTERFACE TO RELAY
 *
 *
 **********************************************************)
(*module Req = Request*)
module DC = Default_cache
let cgDir = ref "ciltrees"
(*******************************************************)
class mySummary = object (self)
  inherit Modsummary.modSumm as super
  val mutable file_opt = (None : Cil.file option)
  val funsum = Hashtbl.create 117
  
  method setFile f = file_opt <- Some f 
  method printStuff (fkey : fKey) = 
   try
      let vi = Cilinfos.getVarinfo fkey in
      log "mySummary object : getMods: %s => "
      (varinfo2str vi)
   with _ ->
      log "mySummary object Exception => "

   method setSummary (fkey:fKey) (s:sum_typ) :unit=
   Hashtbl.replace funsum fkey s

   method getMods  (fkey : fKey) : sum_typ = 
      (* self#printStuff fkey;*)
      match absFind Hashtbl.find funsum fkey  with
      | Some s -> 
            (*log "Found summary (%d)\n" (List.length s); *)
            s 
      | None -> (* log "Summary not found\n"; *)
                []
end
let sum = new mySummary

let doConfig cg = 
let configFile = ref "client.cfg"
in let settings = Config.initSettings !configFile in
   loadFunAlias (); (*local function*)
   DC.makeLCaches (!cgDir);
   Threads.initSettings settings;
   Cilinfos.setDebug true;
   Cilinfos.reloadRanges !cgDir;
  (* RS.initSummaries settings cg;*)
   A.initSettings settings !cgDir; (* do not remove*)
   SS.initSummaries cg;
   SPTR.setErrorFn (fun x-> log "%s" x);
   SPTR.setMallocFn malloc_map;
   SPTR.init settings cg (sum :> Modsummary.modSumm)
(**********************************************************
 *
 *        MAIN PROGRAM
 *
 *
 **********************************************************)

(***************************************************)
(* Commandline handling                            *)

(** File containing pre-processed call graph info *)
let cgFile = ref ""

let setCGFile (fname:string) = 
  cgFile := fname


(* Command-line argument parsing *)
    
let argSpecs = 
  [("-cg", Arg.String setCGFile, "name of call graph file");
   ("-debug", Arg.Unit (fun _ -> enable_log := true), 
    "print debug information")
  ]

let anonArgFun (arg:string) : unit = 
  ()
    
let usageMsg = getUsageString "-cg fname [-debug] [-o] \n"

let simpleBottomUp cg f = 
 let visited = ref IntSet.empty in 
 let worklist = Queue.create () in
 let addw x y = (* add once to worklist*)
  let b = (strstr y.name "__globinit_cil" 0) = -1 in
  if b && not (IntSet.mem x !visited) then
  begin
    log "\nAdding leaf node %s\n" y.name;
    Queue.add (x,y) worklist;
    visited := IntSet.add x !visited 
  end
 in
 let nextw () = Queue.take worklist  in 
 let hasnextw () = not (Queue.is_empty worklist) in  
 log "BOTTOM UP\n";
 FMap.iter 
  (fun fkey node->
     if node.callees = [] then 
        addw fkey node
     else
        log "Function : %s has callees\n" node.name;  
  ) cg;

 while hasnextw () do 
   let (a,b) = nextw () in
   List.iter 
    (fun fkey -> addw fkey (FMap.find fkey cg)) b.callers;
   f a b 
 done

 (*
let mkstack visited = 
 let stack = Stack.create () in
 let push_stack x y = 
  let b = (strstr y.name "__globinit_cil" 0) = -1 in
  if b && not (IntSet.mem x !visited) then
  begin
    log "\nAdding leaf node %s\n" y.name;
    Stack.push (x,y) stack;
    visited := IntSet.add x !visited 
  end
 in
 let nextw () = Stack.pop worklist  in 
 let hasnextw () = not (Stack.is_empty stack) in  

let simpleBottomUp2 cg f = 
 let visited = ref IntSet.empty in 
 let roots =   Callg.getRoots cg in
 

 while not (Stack.is_empty stack) do 

 done;*)

let printScc cg scc = 
   FSet.iter
    (fun fkey ->
       let n = (FMap.find fkey cg) in 
       log "%s " n.name
    ) scc.scc_nodes  



let doBottomUp cg  (f:int->simpleCallN->unit) =
 let visited = ref Fstructs.IntSet.empty in 
 let worklist = Queue.create () in
 let addw x  = Queue.add x worklist in
 let nextw () = Queue.take worklist  in 
 let hasnextw () = not (Queue.is_empty worklist) in  
 let sccCG  = getSCCGraph cg in
 let find_unvisited_callees scc = 
  let fld sccK (map: Fstructs.IntSet.t) = 
    let scc = Fstructs.IntMap.find sccK sccCG in
    try begin 
      (FSet.iter (fun fkey ->
        if not (Fstructs.IntSet.mem fkey !visited) then
         raise Not_found 
      ) scc.scc_nodes);
      map
    end with Not_found -> Fstructs.IntSet.add sccK map
  in Fstructs.IntSet.fold 
    fld scc.scc_callees Fstructs.IntSet.empty
 in

(* let unprocessed_callees scc = 
  let c = IntSet.fold 
in
   ()
 in*)
 (*log "\nLEAF:\n--------------------\n";*)
 Fstructs.IntMap.iter
 (fun sccK scc -> 
      if isLeaf scc then 
         begin
            Queue.add sccK worklist
            (*;printScc cg scc*)
         end
 ) sccCG;

 while hasnextw () do
 (*log "\nQUEUE:\n--------------------\n";
  Queue.iter (fun sccID ->
   let scc = Fstructs.IntMap.find sccID sccCG in
               printScc cg scc) worklist;
  log "\n\n";*)
  let sccK = nextw () in
  try
   (*process strongly connected component*)
   let scc = Fstructs.IntMap.find sccK sccCG in

   let uv = find_unvisited_callees scc in

   if Fstructs.IntSet.is_empty uv then
   begin
      (* process nodes of scc*)
      FSet.iter
       (fun fkey ->
         (* ensure that each key is visited once *)
         if not (Fstructs.IntSet.mem fkey !visited) then
         begin
         begin try
           f fkey (FMap.find fkey cg);
          with
           | Not_found -> 
             rs_impossible "doBottomUp/1" 
          end;
           visited := Fstructs.IntSet.add fkey !visited 
         end
       ) scc.scc_nodes;
       (* add parent sccs to worklist*)
       Fstructs.IntSet.iter addw scc.scc_callers 
   end
   else begin
    Fstructs.IntSet.iter addw uv; (*add unfinished stuff*)
    addw sccK (* process scc once children are done*)
   end
  with 
   | Not_found -> 
     rs_impossible "doBottomUp/0" 
 done 


(**)
let main () = 
(*  try*)
    Arg.parse argSpecs anonArgFun usageMsg;
    
    (* Didn't know how to require the -cg file, etc., so check manually *)
    if (!cgFile = "") then
      begin
        Arg.usage argSpecs usageMsg;
        exit 1
      end
    else
      begin
        Cil.initCIL ();


        let saveFile pre f fname =
         let mkpath  = Filename.concat in 
         let split_path f = 
            (Filename.dirname f, Filename.basename f) in
        (let (_,fname') = split_path f.fileName in
         let fl =  mkpath "/tmp"  (pre ^ fname' )  in
         let oc = open_out  fl  in
         dumpFile defaultCilPrinter oc fl  f;
         close_out oc
        )
        in
      let saveDir dir pre = 
        Filetools.walkDir (saveFile pre) dir; 
      in

        let lkvis = new callVisitor in
        (* Get Callgraph structures *)
        let cg  = readCalls !cgFile in 

        (*let sccCG  = Scc.getSCCGraph cg in

*)
        doConfig cg; 
        (ignore (typ_node_t ()));
        log "-------------------------------------\nDEADLOCK ANALYSIS\n-------------------------------------\n";



        (* Then run symstate on each individual function *)
        let fCache = (new FC.fcache FC.cilLoader 
                        (Filename.dirname !cgFile) 11) in
      
        (*
           FMap.iter 
             (fun fk node ->
             if node.defFile <> ""  then begin
             log "\n\n\n\nSAVING file: %s => " node.defFile;
              let ast = fCache#getFile node.defFile in
              if   (strstr node.defFile "isaac" 0) <> -1 then 
               saveFile "3" ast ast.fileName
             end
             ) cg;*)

       let round = ref 0 in
        (*FMap.iter *)
        doBottomUp cg
          (fun fkey node ->
             if node.defFile <> ""  then begin
             log "\n\n\n\nPROCESSING file: %s => " node.defFile;
             let ast = fCache#getFile node.defFile in

            (sum#setFile) ast;
             setCilFile ast;
             (*registerStuff fCache cg;*)
             (*A.setCurrentFile ast;*)
          (*  dglob (GCompTag  (ci_struct_frame,
                               Cil.locUnknown));
             dglob mk_rstack_gdecl;*)
             match getCFG fkey ast with
               Some (fn) ->
                 log "function: %s\n" 
                  fn.svar.vname;
            round := !round+1;

 
                 let args =  fn.sformals  in 
                 let _ = SPTR.doSymState2 !round fn in 
                 (* ignore ret value for now *)
                 start_time ("doFundec "^fn.svar.vname);

                 let effect  =(lkvis#doFundec fn args) in


                  (sum#setSummary) fkey (lkvis#getMods ());
                  end_time ("doFundec "^fn.svar.vname);
                  start_time ("IntraEffect "^fn.svar.vname);

                  doIntraEffect 
                        (lkvis#simpleEffectOf) fn effect;

                  doInsertOffsets fn;

                  end_time ("IntraEffect "^fn.svar.vname)
             | None ->  ()
             end;
             print_mb "main_loop";
             (*Gc.minor ()*)
          )  ;
          sum_time ();
          (* cg;*)
          (*let main_vid = (lkvis#vidOfMain ()) in*)
          (*make_seq main_vid 3 (-1) 5 6 4; *)
          log "DONE.\n";
          let ret = if hasErrors () then 
                      ( log "Compilation failed.\n"; -1 )
                   else ( (* save instrumented sources *)
                       Unix.mkdir
                       (mkpath "ciltrees" "gen") 0o700;
                       List.iter (fun f ->
                         doFinalize false f) !cilfiles; 
                         log "Compilation OK\n"; 0) in
         (* log "Found it ? %b \n" 
          (opt2bool (varinfo_of_name "dummy___stack"));*)

        exit ret;
      end
;;
main () ;;
